import json
import os
import warnings
import torch  # to handle devices

from leaspy.io.settings import algo_default_data_dir
from leaspy.io.settings.outputs_settings import OutputsSettings
from leaspy.algo.algo_factory import AlgoFactory
from leaspy.algo.utils.algo_with_device import AlgoWithDeviceMixin

from leaspy.exceptions import LeaspyAlgoInputError
from leaspy.utils.typing import KwargsType, Optional


class AlgorithmSettings:
    """
    Used to set the algorithms' settings.

    All parameters, except the choice of the algorithm, is set by default.
    The user can overwrite all default settings.

    Parameters
    ----------
    name : str
        The algorithm's name. Must be in:
            * For `fit` algorithms:
                * ``'mcmc_saem'``
                * ``'lme_fit'`` (for LME model only)
            * For `personalize` algorithms:
                * ``'scipy_minimize'``
                * ``'mean_real'``
                * ``'mode_real'``
                * ``'constant_prediction'`` (for constant model only)
                * ``'lme_personalize'`` (for LME model only)
            * For `simulate` algorithms:
                * ``'simulation'``

    **kwargs : any
        Depending on the algorithm you are setting up, various parameters are possible (not exhaustive):
            * seed : int, optional, default None
                Used for stochastic algorithms.
            * model_initialization_method : str, optional
                For **fit** algorithms only, give a model initialization method,
                according to those possible in :func:`~.models.utils.initialization.model_initialization.initialize_parameters`.
            * algo_initialization_method : str, optional
                Personalize the algorithm initialization method,
                according to those possible for the given algorithm (refer to its documentation in :mod:`leaspy.algo`).
            * n_iter : int, optional
                Number of iteration. There is no stopping criteria for the all the MCMC SAEM algorithms.
            * n_burn_in_iter : int, optional
                Number of iteration during burning phase, used for the MCMC SAEM algorithms.
            * use_jacobian : bool, optional, default True
                Used in ``scipy_minimize`` algorithm to perform a `L-BFGS` instead of a `Powell` algorithm.
            * n_jobs : int, optional, default 1
                Used in ``scipy_minimize`` algorithm to accelerate calculation with parallel derivation using joblib.
            * progress_bar : bool, optional, default True
                Used to display a progress bar during computation.
            * device: str or torch.device, optional
                Specifies on which device the algorithm will run.
                Only 'cpu' and 'cuda' are supported for this argument.
                Only ``'mcmc_saem'``, ``'mean_real'`` and ``'mode_real'`` algorithms support this setting.

        For the complete list of the available parameters for a given algorithm, please directly refer to its documentation.

    Attributes
    ----------
    name : str
        The algorithm's name.
    model_initialization_method : str, optional
        For fit algorithms, give a model initialization method,
        according to those possible in :func:`~.models.utils.initialization.model_initialization.initialize_parameters`.
    algo_initialization_method : str, optional
        Personalize the algorithm initialization method,
        according to those possible for the given algorithm (refer to its documentation in :mod:`leaspy.algo`).
    seed : int, optional, default None
        Used for stochastic algorithms.
    parameters : dict
        Contains the other parameters: `n_iter`, `n_burn_in_iter`, `use_jacobian`, `n_jobs` & `progress_bar`.
    logs : :class:`.OutputsSettings`, optional
        Used to create a ``logs`` file during a model calibration containing convergence information.
    device : str (or torch.device), optional, default 'cpu'
        Used to specify on which device the algorithm will run. This should either be:
        'cpu' or 'cuda' and is only supported in specific algorithms (inheriting `AlgoWithDeviceMixin`).
        Note that specifying an indexed CUDA device (such as 'cuda:1') is not supported.
        In order to specify the precise cuda device index, one should use the `CUDA_VISIBLE_DEVICES` environment variable.

    Raises
    ------
    :exc:`.LeaspyAlgoInputError`

    See Also
    --------
    :mod:`leaspy.algo`

    Notes
    -----
    For developers: use ``_dynamic_default_parameters`` to dynamically set some default parameters,
    depending on other parameters that were set, while these `dynamic` parameters were not set.

    Example:
        you could want to set burn in iterations or annealing iterations
        as fractions of non-default number of iterations given.

    Format:

    ::

        {algo_name: [
            (functional_condition_to_trigger_dynamic_setting(kwargs),
            {
                nested_keys_of_dynamic_setting: dynamic_value(kwargs)
            })
        ]}
    """

    # TODO should be in the each algo class directly?
    _dynamic_default_parameters = {
        'lme_fit': [
            (
                lambda kw: 'force_independent_random_effects' in kw and kw['force_independent_random_effects'],
                {
                    ('method',): lambda kw: ['lbfgs','bfgs'] # Powell & Nelder-Mead methods cannot ensure respect of "free"
                }
            )
        ]
    }

    # known keys for all algorithms (<!> not all of them are mandatory!)
    _known_keys = ['name', 'seed', 'algorithm_initialization_method', 'model_initialization_method', 'parameters', 'device'] # 'logs' are not handled in exported files

    def __init__(self, name: str, **kwargs):

        self.name = name
        self.parameters: KwargsType = None # {}
        self.seed: Optional[int] = None
        self.algorithm_initialization_method: str = None # Initialization of the algorithm itself
        self.model_initialization_method: str = None # Initialization of the model parameters (independently of the algorithm, for fit family only)
        self.logs = None

        default_algo_settings_path = os.path.join(algo_default_data_dir, 'default_' + name + '.json')

        if os.path.isfile(default_algo_settings_path):
            self._load_default_values(default_algo_settings_path)
        else:
            raise LeaspyAlgoInputError(f"The algorithm name '{name}' you provided does not exist")

        self._manage_kwargs(kwargs)
        self.check_consistency()

    @property
    def algo_class(self):
        """Class of the algorithm derived from its name (shorthand)."""
        return AlgoFactory.get_class(self.name)

    def check_consistency(self) -> None:
        """
        Check internal consistency of algorithm settings and warn or raise a `LeaspyAlgoInputError` if not.
        """
        algo_family = self.algo_class.family
        if self.model_initialization_method is not None and algo_family != 'fit':
            warnings.warn('You should not define `model_initialization_method` in your algorithm: '
                          f'"{self.name}" is not a `fit` algorithm so it has no effect.')

        if self.seed is not None and self.algo_class.deterministic:
            warnings.warn(f'You can skip defining `seed` since the algorithm {self.name} is deterministic.')

        if hasattr(self, 'device') and not issubclass(self.algo_class, AlgoWithDeviceMixin):
            warnings.warn(f'The algorithm "{self.name}" does not support user-specified devices (this '
                           'is supported only for specific algorithms) and will use the default device (CPU).')

    @classmethod
    def _recursive_merge_dict_warn_extra_keys(cls, ref: dict, new: dict, *, prefix_keys: str = ''):
        """Merge in-place dictionary `ref` with the values from `new`, for dict keys, merge is recursive."""
        extra_keys = [prefix_keys + k for k in new if k not in ref]
        if extra_keys:
            warnings.warn(f'The parameters {extra_keys} were not present by default and are likely to be unsupported.')
        for k, v in new.items():
            if k not in ref or not isinstance(ref[k], dict):
                ref[k] = v
            else:
                if not isinstance(v, dict):
                    raise LeaspyAlgoInputError(f"Algorithm parameter `{prefix_keys + k}` should be a dictionary, not '{v}' of type {type(v)}.")
                cls._recursive_merge_dict_warn_extra_keys(ref[k], v, prefix_keys=f'{prefix_keys}{k}.')

    @classmethod
    def load(cls, path_to_algorithm_settings: str):
        """
        Instantiate a AlgorithmSettings object a from json file.

        Parameters
        ----------
        path_to_algorithm_settings : str
            Path of the json file.

        Returns
        -------
        :class:`.AlgorithmSettings`
            An instanced of AlgorithmSettings with specified parameters.

        Raises
        ------
        :exc:`.LeaspyAlgoInputError`
            if anything is invalid in algo settings

        Examples
        --------
        >>> from leaspy import AlgorithmSettings
        >>> leaspy_univariate = AlgorithmSettings.load('outputs/leaspy-univariate_model-settings.json')
        """
        with open(path_to_algorithm_settings) as fp:
            settings = json.load(fp)

        if 'name' not in settings.keys():
            raise LeaspyAlgoInputError("Your json file must contain a 'name' attribute!")

        algorithm_settings = cls(settings['name'])

        if 'parameters' in settings.keys():
            print("You overwrote the algorithm default parameters")
            cls._recursive_merge_dict_warn_extra_keys(algorithm_settings.parameters, cls._get_parameters(settings))

        if 'seed' in settings.keys():
            print("You overwrote the algorithm default seed")
            algorithm_settings.seed = cls._get_seed(settings)

        if 'algorithm_initialization_method' in settings.keys():
            print("You overwrote the algorithm default initialization method")
            algorithm_settings.algorithm_initialization_method = cls._get_algorithm_initialization_method(settings)

        if 'model_initialization_method' in settings.keys():
            print("You overwrote the model default initialization method")
            algorithm_settings.model_initialization_method = cls._get_model_initialization_method(settings)

        if 'device' in settings.keys():
            print("You overwrote the algorithm default device")
            algorithm_settings.device = cls._get_device(settings)

        if 'loss' in settings.keys():
            raise LeaspyAlgoInputError('`loss` keyword for AlgorithmSettings is not supported any more. '
                                       'Please define `noise_model` directly in your Leaspy model.')

        # Unknown keys
        # TODO: this class should really be refactored so not to copy in 3 methods same stuff (manage_kwargs, load & _check_default_settings)
        unknown_keys = set(settings.keys()).difference(cls._known_keys)
        if unknown_keys:
            raise LeaspyAlgoInputError(f'Unexpected keys {unknown_keys} in algorithm settings.')

        algorithm_settings.check_consistency()

        return algorithm_settings

    def save(self, path: str, **kwargs):
        """
        Save an AlgorithmSettings object in a json file.

        TODO? save leaspy version as well for retro/future-compatibility issues?

        Parameters
        ----------
        path : str
            Path to store the AlgorithmSettings.
        **kwargs
            Keyword arguments for json.dump method.
            Default: dict(indent=2)

        Examples
        --------
        >>> from leaspy import AlgorithmSettings
        >>> settings = AlgorithmSettings('scipy_minimize', seed=42)
        >>> settings.save('outputs/scipy_minimize-settings.json')
        """
        json_settings = {
            "name": self.name,
            "seed": self.seed,
            "algorithm_initialization_method": self.algorithm_initialization_method,
        }

        algo_family = self.algo_class.family
        if algo_family == 'fit':
            json_settings['model_initialization_method'] = self.model_initialization_method
            json_settings['device'] = self.device

        """
        TODO: save config of logging as well (OutputSettings needs to be JSON serializable...)
        if self.logs is not None:
            json_settings['logs'] = self.logs
        """

        # append parameters key after "hyperparameters"
        json_settings['parameters'] = self.parameters

        # Default json.dump kwargs:
        kwargs = {'indent': 2, **kwargs}

        with open(os.path.join(path), "w") as json_file:
            json.dump(json_settings, json_file, **kwargs)

    def set_logs(self, path: Optional[str] = None, **kwargs):
        """
        Use this method to monitor the convergence of a model calibration.

        It create graphs and csv files of the values of the population parameters (fixed effects) during the calibration

        Parameters
        ----------
        path : str, optional
            The path of the folder to store the graphs and csv files.
            No data will be saved if it is None, as well as save_periodicity and plot_periodicity.
        **kwargs
            * console_print_periodicity: int, optional, default 100
                Display logs in the console/terminal every N iterations.
            * save_periodicity: int, optional, default 50
                Saves the values in csv files every N iterations.
            * plot_periodicity: int, optional, default 1000
                Generates plots from saved values every N iterations.
                Note that:
                    * it should be a multiple of save_periodicity
                    * setting a too low value (frequent) we seriously slow down you calibration
            * overwrite_logs_folder: bool, optional, default False
                Set it to ``True`` to overwrite the content of the folder in ``path``.

        Raises
        ------
        :exc:`.LeaspyAlgoInputError`
            If the folder given in ``path`` already exists and if ``overwrite_logs_folder`` is set to ``False``.

        Notes
        -----
        By default, if the folder given in ``path`` already exists, the method will raise an error.
        To overwrite the content of the folder, set ``overwrite_logs_folder`` it to ``True``.
        """
        # TODO: all this logic should be delegated in dedicated OutputSettings class...!

        settings = {
            'path': path,
            'console_print_periodicity': 100,
            'save_periodicity': 50,
            'plot_periodicity': 1000,
            'overwrite_logs_folder': False
        }

        for k, v in kwargs.items():
            if k in ['console_print_periodicity', 'plot_periodicity', 'save_periodicity']:
                if v is not None and not isinstance(v, int):
                    raise LeaspyAlgoInputError(f'You must provide a integer to the input <{k}>! '
                                    f'You provide {v} of type {type(v)}.')
                settings[k] = v
            elif k in ['overwrite_logs_folder']:
                if not isinstance(v, bool):
                    raise LeaspyAlgoInputError(f'You must provide a boolean to the input <{k}>! '
                                    f'You provide {v} of type {type(v)}.')
                settings[k] = v
            else:
                warnings.warn(f"The kwarg '{k}' you provided is not valid and was skipped.")

        self.logs = OutputsSettings(settings)

    def _manage_kwargs(self, kwargs):

        _special_kwargs = {
            'seed': self._get_seed,
            'algorithm_initialization_method': self._get_algorithm_initialization_method,
            'model_initialization_method': self._get_model_initialization_method,
            'device': self._get_device,
        }

        for k, v in kwargs.items():
            if k in _special_kwargs:
                k_getter = _special_kwargs[k]
                setattr(self, k, k_getter(kwargs))

        kwargs_interpreted_as_parameters = {k: v for k, v in kwargs.items() if k not in _special_kwargs}
        self._recursive_merge_dict_warn_extra_keys(self.parameters, kwargs_interpreted_as_parameters)

        # dynamic default parameters
        if self.name in self._dynamic_default_parameters:

            for func_condition, associated_defaults in self._dynamic_default_parameters[self.name]:

                if not func_condition(kwargs):
                    continue

                # loop on dynamic defaults
                for nested_levels, val_getter in associated_defaults.items():
                    # check that the dynamic default that we want to set is not already overwritten
                    if self._get_nested_dict(kwargs, nested_levels) is None:
                        self._set_nested_dict(self.parameters, nested_levels, val_getter(kwargs))

    @staticmethod
    def _get_nested_dict(nested_dict: dict, nested_levels, default = None):
        """
        Get a nested key of a dict or default if any previous level is missing.

        Examples
        --------
        >>> _get_nested_dict(d, ('a','b'), -1) == ...
            * -1 if 'a' not in d
            * -1 if 'b' not in d['a']
            * d['a']['b'] else

        >>> _get_nested_dict(d, (), ...) == d
        """
        it_levels = iter(nested_levels)

        while isinstance(nested_dict, dict):
            try:
                next_lvl = next(it_levels)
            except StopIteration:
                break

            # get next level dict
            nested_dict = nested_dict.get(next_lvl, default)

        return nested_dict

    @classmethod
    def _set_nested_dict(cls, nested_dict: dict, nested_levels, val):
        """
        Set a nested key of a dict.
        Precondition: all intermediate levels must exist.
        """
        *nested_top_levels, last_level = nested_levels
        dict_to_set = cls._get_nested_dict(nested_dict, nested_top_levels, default=None)
        assert isinstance(dict_to_set, dict)
        dict_to_set[last_level] = val # inplace

    def _load_default_values(self, path_to_algorithm_settings):

        with open(path_to_algorithm_settings) as fp:
            settings = json.load(fp)

        self._check_default_settings(settings)
        # TODO: Urgent => The following function should in fact be algorithm-name specific!! As for the constant prediction
        # Etienne: I'd advocate for putting all non-generic / parametric stuff in special methods / attributes of corresponding algos... so that everything is generic here
        # Igor : Agreed. This class became a real mess.

        self.name = self._get_name(settings)
        self.parameters = self._get_parameters(settings)

        self.algorithm_initialization_method = self._get_algorithm_initialization_method(settings)

        # optional hyperparameters depending on type of algorithm
        algo_class = self.algo_class

        if not algo_class.deterministic:
            self.seed = self._get_seed(settings)

        if algo_class.family == 'fit':
            self.model_initialization_method = self._get_model_initialization_method(settings)

        if 'device' in settings:
            self.device = self._get_device(settings)

    @classmethod
    def _check_default_settings(cls, settings):

        # Unknown keys
        unknown_keys = set(settings.keys()).difference(cls._known_keys)
        if unknown_keys:
            raise LeaspyAlgoInputError(f'Unexpected keys {unknown_keys} in algorithm settings.')

        # Missing keys
        error_tpl = "The '{}' key is missing in the algorithm settings (JSON file) you are loading."

        for mandatory_key in ['name', 'parameters']:
            if mandatory_key not in settings.keys():
                raise LeaspyAlgoInputError(error_tpl.format(mandatory_key))

        algo_class = AlgoFactory.get_class(settings['name'])

        if not algo_class.deterministic and 'seed' not in settings.keys():
            raise LeaspyAlgoInputError(error_tpl.format('seed'))

        if 'algorithm_initialization_method' not in settings.keys():
            raise LeaspyAlgoInputError(error_tpl.format('algorithm_initialization_method'))

        if algo_class.family == 'fit' and 'model_initialization_method' not in settings.keys():
            raise LeaspyAlgoInputError(error_tpl.format('model_initialization_method'))

    @staticmethod
    def _get_name(settings):
        return settings['name'].lower()

    @staticmethod
    def _get_parameters(settings):
        return settings['parameters']

    @staticmethod
    def _get_seed(settings):
        if settings['seed'] is None:
            return None

        try:
            return int(settings['seed'])
        except Exception:
            warnings.warn(f"The 'seed' parameter you provided ({settings['seed']}) cannot be converted to int, using None instead.")
            return None

    @staticmethod
    def _get_algorithm_initialization_method(settings):
        if settings['algorithm_initialization_method'] is None:
            return None
        # TODO : There should be a list of possible initialization method. It can also be discussed depending on the algorithms name
        return settings['algorithm_initialization_method']

    @staticmethod
    def _get_model_initialization_method(settings):
        if settings['model_initialization_method'] is None:
            return None
        # TODO : There should be a list of possible initialization method. It can also be discussed depending on the algorithms name
        return settings['model_initialization_method']

    @staticmethod
    def _get_device(settings):
        # in case where a torch.device object was used, we convert it to the
        # corresponding string (torch.device('cuda') is converted into 'cuda')
        # in order for the AlgorithmSettings to be saved into json files if needed
        if isinstance(settings['device'], torch.device):
            return settings['device'].type

        # getting the type of torch.device(...) allows to convert 'cuda:2' to 'cuda'
        # which prevents potential issues when using torch.set_default_tensor_type
        return torch.device(settings['device']).type



import os

import pandas as pd

from leaspy import Leaspy, IndividualParameters

dataset_all = ['alzheimer-multivariate', 'parkinson-multivariate', 'parkinson-putamen',
               'parkinson-putamen-train_and_test']
model_all = ['alzheimer-multivariate', 'parkinson-multivariate', 'parkinson-putamen-train']
ip_all = ['alzheimer-multivariate', 'parkinson-multivariate', 'parkinson-putamen-train']

module_path = os.path.dirname(__file__)
data_paths = {dataset_name: os.path.join(module_path, 'data', dataset_name + '.csv') for dataset_name in dataset_all}
model_paths = {model_name: os.path.join(module_path, 'model_parameters', model_name + '-model_parameters.json') for model_name in model_all}
ip_paths = {ip_name: os.path.join(module_path, 'individual_parameters', ip_name + '-individual_parameters.csv') for ip_name in ip_all}

# TODO: add some leaspy_instance
# TODO: add some individual_parameters


class Loader:
    """
    Contains static methods to load synthetic longitudinal dataset, calibrated :class:`.Leaspy` instances &
    :class:`.IndividualParameters`.

    Attributes
    ----------
    data_paths : dict [str, str]
        Contains the datasets' names and their respective path within ``leaspy.datasets`` subpackage.
    model_paths : dict [str, str]
        Contains the `Leaspy` instances' names and their respective path within ``leaspy.datasets`` subpackage.
    ip_paths : dict [str, str]
        Contains the individual parameters' names and their respective path within ``leaspy.datasets`` subpackage.

    Notes
    -----
    * A `Leaspy` instance named ``<name>`` have been calibrated on the dataset ``<name>``.
    * An `IndividualParameters` name ``<name>`` have been computed by personalizing the `Leaspy` instance
      named ``<name>`` on the dataset ``<name>``.

    See the documentation of each method to get their respective available names.
    """

    def __init__(self):
        self.data_paths = data_paths
        self.model_paths = model_paths
        self.ip_paths = ip_paths

    @staticmethod
    def load_dataset(dataset_name):
        """
        Load synthetic longitudinal observations mimicking cohort of subjects with neurodegenerative disorders.

        Parameters
        ----------
        dataset_name : {'parkinson-multivariate', 'alzheimer-multivariate', 'parkinson-putamen', 'parkinson-putamen-train_and_test'}
            Name of the dataset.

        Returns
        -------
        :class:`pandas.DataFrame`
            DataFrame containing the IDs, timepoints and observations.

        Notes
        -----
        All `DataFrames` have the same structures.

        * Index: a :class:`pandas.MultiIndex` - ``['ID', 'TIME']`` which contain IDs and timepoints. The `DataFrame` is sorted by index.
          So, one line corresponds to one visit for one subject. The `DataFrame` having `'train_and_test'` in their name
          also have ``'SPLIT'`` as the third index level. It differentiate `train` and `test` data.

        * Columns: One column correspond to one feature (or score).
        """
        df = pd.read_csv(data_paths[dataset_name], dtype={'ID': str})
        if 'SPLIT' in df.columns:
            df.set_index(['ID', 'TIME', 'SPLIT'], inplace=True)
        else:
            df.set_index(['ID', 'TIME'], inplace=True)
        return df.sort_index()

    @staticmethod
    def load_leaspy_instance(instance_name):
        """
        Load a Leaspy instance with a model already calibrated on the synthetic dataset corresponding to the name
        of the instance.

        Parameters
        ----------
        instance_name : {'alzheimer-multivariate', 'parkinson-multivariate', 'parkinson-putamen-train'}
            Name of the instance.

        Returns
        -------
        :class:`.Leaspy`
            Leaspy instance with a model already calibrated.
        """
        return Leaspy.load(model_paths[instance_name])

    @staticmethod
    def load_individual_parameters(ip_name):
        """
        Load a Leaspy instance with a model already calibrated on the synthetic dataset corresponding to the name
        of the instance.

        Parameters
        ----------
        ip_name : {'alzheimer-multivariate', 'parkinson-multivariate', 'parkinson-putamen-train'}
            Name of the individual parameters.

        Returns
        -------
        :class:`.IndividualParameters`
            Leaspy instance with a model already calibrated.
        """
        return IndividualParameters.load(ip_paths[ip_name])



from __future__ import annotations
from typing import TYPE_CHECKING

from random import shuffle
from copy import copy
from abc import abstractmethod

import torch

from leaspy.algo.personalize.abstract_personalize_algo import AbstractPersonalizeAlgo
from leaspy.algo.utils.samplers import AlgoWithSamplersMixin
from leaspy.algo.utils.algo_with_device import AlgoWithDeviceMixin
from leaspy.algo.utils.algo_with_annealing import AlgoWithAnnealingMixin
from leaspy.io.outputs.individual_parameters import IndividualParameters
from leaspy.utils.typing import DictParamsTorch

if TYPE_CHECKING:
    from leaspy.models.abstract_model import AbstractModel
    from leaspy.io.data.dataset import Dataset


class AbstractMCMCPersonalizeAlgo(AlgoWithAnnealingMixin, AlgoWithSamplersMixin, AlgoWithDeviceMixin, AbstractPersonalizeAlgo):
    """
    Base class for MCMC-based personalization algorithms.

    Individual parameters are derived from realizations of individual variables of the model.

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        Settings of the algorithm.
    """

    @abstractmethod
    def _compute_individual_parameters_from_samples_torch(self,
            realizations: DictParamsTorch,
            attachments: torch.FloatTensor,
            regularities: torch.FloatTensor) -> DictParamsTorch:
        """
        Compute dictionary of individual parameters from stacked realizations, attachments and regularities.

        Parameters
        ----------
        realizations : dict[ind_var_name: str, `torch.FloatTensor` of shape (n_iter, n_individuals, *ind_var.shape)]
            The stacked history of realizations for individual variables.
        attachments : `torch.FloatTensor` of shape (n_iter, n_individuals)
            The stacked history of attachments.
        regularities : `torch.FloatTensor` of shape (n_iter, n_individuals)
            The stacked history of regularities (sum on all individual variables / dimensions).

        Returns
        -------
        dict[ind_var_name: str, `torch.FloatTensor` of shape (n_individuals, *ind_var.shape)]
        """

    def _get_individual_parameters(self, model: AbstractModel, dataset: Dataset):

        # Initialize realizations storage object
        realizations_history = []
        attachment_history = []
        regularity_history = []

        # We are not in a calibration any more so attribute_type=None (NOT using MCMC toolbox, since it is undefined!)
        computation_kws = dict(attribute_type=None)

        with self._device_manager(model, dataset):
            # Initialize samplers
            self._initialize_samplers(model, dataset)

            # Initialize Annealing
            self._initialize_annealing()

            # Initialize realizations, only for individual variables, and at their mean values
            pop_var_skipper = lambda info_var: info_var['type'] == 'population'
            realizations = model.initialize_realizations_for_model(
                                    dataset.n_individuals,
                                    skip_variable=pop_var_skipper,
                                    individual_variable_init_at_mean=True,
            )
            ind_vars_names = copy(realizations.reals_ind_variable_names)

            n_iter = self.algo_parameters['n_iter']
            if self.algo_parameters.get('progress_bar', True):
                self._display_progress_bar(-1, n_iter, suffix='iterations')

            # Gibbs sample `n_iter` times (only individual parameters)
            for self.current_iteration in range(1, n_iter+1):

                if self.random_order_variables:
                    shuffle(ind_vars_names)  # shuffle in-place!

                last_attachment = None
                tot_regularities = 0.
                for ind_var_name in ind_vars_names:
                    last_attachment, regularity_var = self.samplers[ind_var_name].sample(dataset, model, realizations,
                                                                                         self.temperature_inv, **computation_kws)
                    tot_regularities += regularity_var

                # Append current realizations if "burn-in phase" is finished
                if not self._is_burn_in():
                    realizations_history.append(realizations.clone_realizations())
                    attachment_history.append(last_attachment.clone().detach())
                    regularity_history.append(tot_regularities.clone().detach())

                # Annealing
                self._update_temperature()

                # TODO? print(self) periodically? or refact OutputManager for not fit algorithms...

                if self.algo_parameters.get('progress_bar', True):
                    self._display_progress_bar(self.current_iteration - 1, n_iter, suffix='iterations')

            # Stack tensor realizations as well as attachments and tot_regularities
            torch_realizations = {
                ind_var_name: torch.stack([realizations[ind_var_name].tensor_realizations
                                           for realizations in realizations_history])
                for ind_var_name in realizations.reals_ind_variable_names
            }
            torch_attachments = torch.stack(attachment_history)
            torch_tot_regularities = torch.stack(regularity_history)

            # Derive individual parameters from `realizations_history` list
            individual_parameters_torch = self._compute_individual_parameters_from_samples_torch(
                                                    torch_realizations, torch_attachments, torch_tot_regularities)

            # Create the IndividualParameters object
            return IndividualParameters.from_pytorch(dataset.indices, individual_parameters_torch)



import csv
import os
import time

from leaspy.io.logs.visualization.plotter import Plotter


class FitOutputManager:
    """
    Class used by :class:`.AbstractAlgo` (and its child classes) to display & save plots and statistics during algorithm execution.

    Parameters
    ----------
    outputs : :class:`~.io.settings.outputs_settings.OutputsSettings`
        Initialize the `FitOutputManager` class attributes, like the logs paths, the console print periodicity and so forth.

    Attributes
    ----------
    path_output : str
        Path of the folder containing all the outputs
    path_plot : str
        Path of the subfolder of path_output containing the logs plots
    path_plot_convergence_model_parameters_1 : str
        Path of the first plot of the convergence of the model's parameters (in the subfolder path_plot)
    path_plot_convergence_model_parameters_2 : str
        Path of the second plot of the convergence of the model's parameters (in the subfolder path_plot)
    path_plot_patients : str
        Path of the subfolder of path_plot containing the plot of the reconstruction of the patients' longitudinal
        trajectory by the model
    path_save_model_parameters_convergence : str
        Path of the subfolder of path_output containing the progression of the model's parameters convergence
    periodicity_plot : int (default 100)
        Set the frequency of the display of the plots
    periodicity_print : int
        Set the frequency of the display of the statistics
    periodicity_save : int
        Set the frequency of the saves of the model's parameters & the realizations
    plot_options : dict
        Contain all the additional information (for now contain only the number of displayed patients by the method
        plot_patient_reconstructions - which is 5 by default)
    plotter : :class:`~.utils.logs.visualization.plotter.Plotter`
        class object used to call visualization methods
    time : float
        Last timestamp (to display the duration between two visualization prints)
    """

    def __init__(self, outputs):

        self.periodicity_print = outputs.console_print_periodicity
        self.periodicity_save = outputs.save_periodicity
        self.periodicity_plot = outputs.plot_periodicity

        self.path_output = outputs.root_path
        self.path_plot = outputs.plot_path
        self.path_plot_patients = outputs.patients_plot_path
        self.path_save_model_parameters_convergence = outputs.parameter_convergence_path
        self.path_plot_convergence_model_parameters_1 = None
        self.path_plot_convergence_model_parameters_2 = None

        if outputs.patients_plot_path is not None:
            self.path_plot_convergence_model_parameters_1 = os.path.join(outputs.plot_path, "convergence_1.pdf")
            self.path_plot_convergence_model_parameters_2 = os.path.join(outputs.plot_path, "convergence_2.pdf")

        # Options
        # TODO : Maybe add to the outputs reader
        # <!> We would need the attributes of the model if we would want to reconstruct values without MCMC toolbox
        # This is the only location where we could need it during the calibration....
        self.plot_options = {'max_patient_number': 5, 'attribute_type': 'MCMC'}
        self.plotter = Plotter()
        self.plotter._show = False  # do not show any of the plots!

        self.time = time.time()

        self.save_last_n_realizations = 100

    def iteration(self, algo, data, model, realizations):
        """
        Call methods to save state of the running computation, display statistics & plots if the current iteration
        is a multiple of `periodicity_print`, `periodicity_plot` or `periodicity_save`

        Parameters
        ----------
        algo : :class:`.AbstractAlgo`
            The running algorithm
        data : :class:`.Data`
            The data used by the computation
        model : :class:`~.models.abstract_model.AbstractModel`
            The model used by the computation
        realizations : :class:`~.io.realizations.collection_realization.CollectionRealization`
            Current state of the realizations
        """

        # <!> only `current_iteration` defined for AbstractFitAlgorithm... TODO -> generalize where possible?
        if not hasattr(algo, 'current_iteration'):
            # emit a warning?
            return

        iteration = algo.current_iteration

        if self.periodicity_print is not None:
            if iteration == 1 or iteration % self.periodicity_print == 0:
                # print first iteration
                print() # newline
                self.print_algo_statistics(algo)
                print()
                self.print_model_statistics(model)
                print()
                self.print_time()

        if self.path_output is None:
            return

        if self.periodicity_save is not None:
            if iteration == 1 or iteration % self.periodicity_save == 0:
                # save first iteration
                self.save_model_parameters_convergence(iteration, model)
                # model.save(...)

        if self.periodicity_plot is not None:
            if iteration % self.periodicity_plot == 0:
                # do not plot first iteration (useless, no lines yet)
                self.plot_patient_reconstructions(iteration, data, model, realizations)
                self.plot_convergence_model_parameters(model)

        if (algo.algo_parameters['n_iter'] - iteration) < self.save_last_n_realizations:
            self.save_realizations(iteration, realizations)

    ########
    ## Printing methods
    ########

    def print_time(self):
        """
        Display the duration since the last print
        """
        current_time = time.time()
        print(f"Duration since last print: {current_time - self.time:.3f}s")
        self.time = current_time

    def print_model_statistics(self, model):
        """
        Print model's statistics

        Parameters
        ----------
        model : :class:`~.models.abstract_model.AbstractModel`
            The model used by the computation
        """
        print(model)

    def print_algo_statistics(self, algo):
        """
        Print algorithm's statistics

        Parameters
        ----------
        algo : :class:`.AbstractAlgo`
            The running algorithm
        """
        print(algo)

    ########
    ## Saving methods
    ########

    def save_model_parameters_convergence(self, iteration, model):
        """
        Save the current state of the model's parameters

        Parameters
        ----------
        iteration : int
            The current iteration
        model : :class:`~.models.abstract_model.AbstractModel`
            The model used by the computation
        """
        model_parameters = model.parameters

        # TODO maybe better way ???
        model_parameters_save = model_parameters.copy()

        # TODO I Stopped here, 2d array saves should be fixed.

        # Transform the types
        for key, value in model_parameters.items():

            if value.ndim > 1:
                if key == "betas":
                    model_parameters_save.pop(key)
                    for column in range(value.shape[1]):
                        model_parameters_save[f"{key}_{column}"] = value[:, column].tolist()
                # P0, V0
                elif value.shape[0] == 1 and len(value.shape) > 1:
                    model_parameters_save[key] = value[0].tolist()
            elif value.ndim == 1:
                model_parameters_save[key] = value.tolist()
            else:  # ndim == 0
                model_parameters_save[key] = [value.tolist()]

        # Save the dictionary
        for key, value in model_parameters_save.items():
            path = os.path.join(self.path_save_model_parameters_convergence, key + ".csv")
            with open(path, 'a', newline='') as filename:
                writer = csv.writer(filename)
                writer.writerow([iteration] + value)

    def save_realizations(self, iteration, realizations):
        """
        Save the current realizations. The path is given by the attribute path_save_model_parameters_convergence

        Parameters
        ----------
        iteration : int
            The current iteration
        realizations : :class:`~.io.realizations.collection_realization.CollectionRealization`
            Current state of the realizations
        """
        # TODO: not generic at all
        for name in ['xi', 'tau']:
            value = realizations[name].tensor_realizations.squeeze(1).detach().tolist()
            path = os.path.join(self.path_save_model_parameters_convergence, name + ".csv")
            with open(path, 'a', newline='') as filename:
                writer = csv.writer(filename)
                # writer.writerow([iteration]+list(model_parameters.values()))
                writer.writerow([iteration] + value)
        if "sources" in realizations.reals_ind_variable_names:
            for i in range(realizations['sources'].tensor_realizations.shape[1]):
                value = realizations['sources'].tensor_realizations[:, i].detach().tolist()
                path = os.path.join(self.path_save_model_parameters_convergence, 'sources' + str(i) + ".csv")
                with open(path, 'a', newline='') as filename:
                    writer = csv.writer(filename)
                    # writer.writerow([iteration]+list(model_parameters.values()))
                    writer.writerow([iteration] + value)

    ########
    ## Plotting methods
    ########

    def plot_convergence_model_parameters(self, model):
        """
        Plot the convergence of the model parameters (calling the `Plotter`)

        Parameters
        ----------
        model : :class:`~.models.abstract_model.AbstractModel`
            The model used by the computation
        """
        self.plotter.plot_convergence_model_parameters(self.path_save_model_parameters_convergence,
                                                       self.path_plot_convergence_model_parameters_1,
                                                       self.path_plot_convergence_model_parameters_2,
                                                       model)

    #def plot_model_average_trajectory(self, model):
    #    raise NotImplementedError

    def plot_patient_reconstructions(self, iteration, data, model, realizations):
        """
        Plot on the same graph for several patients their real longitudinal values and their reconstructions by the model

        Parameters
        ----------
        iteration : int
            The current iteration
        data : :class:`.Data`
            The data used by the computation
        model : :class:`~.models.abstract_model.AbstractModel`
            The model used by the computation
        realizations : :class:`~.io.realizations.collection_realization.CollectionRealization`
            Current state of the realizations
        """
        path_iteration = os.path.join(self.path_plot_patients, f'plot_patients_{iteration}.pdf')
        param_ind = model.get_param_from_real(realizations)
        self.plotter.plot_patient_reconstructions(path_iteration, data, model, param_ind, **self.plot_options)

        """
        colors = cm.rainbow(np.linspace(0, 1, self.plot_options['maximum_patient_number']+2))
        reals_pop, reals_ind = realizations

        fig, ax = plt.subplots(1, 1)

        for i, idx in enumerate(data.indices):
            model_value = model.compute_individual(data[idx], reals_pop, reals_ind[idx])
            score = data[idx].tensor_observations
            ax.plot(data[idx].tensor_timepoints.detach().numpy(), model_value.detach().numpy(), c=colors[i])
            ax.plot(data[idx].tensor_timepoints.detach().numpy(), score.detach().numpy(), c=colors[i], linestyle='--',
                    marker='o')

            if i > self.plot_options['maximum_patient_number']:
                break

        # Plot average model
        tensor_timepoints = torch.tensor(np.linspace(data.time_min, data.time_max, 40).reshape(-1,1), dtype=torch.float32)
        model_average = model.compute_average(tensor_timepoints)
        ax.plot(tensor_timepoints.detach().numpy(), model_average.detach().numpy(), c='black', linewidth=4, alpha=0.3)

        plt.savefig(os.path.join(self.path_plot_patients,f'plot_patients_{iteration}.pdf'))
        plt.close()
        """




from typing import Callable
from functools import wraps

def suffixed_method(f: Callable):
    """Decorator function to delegate a method to the subtype (suffixed) method if it exists."""
    @wraps(f)
    def f_sub(self, *args, **kws):
        assert self._subtype_suffix is not None and len(self._subtype_suffix) > 0
        f_suffixed = f.__name__ + self._subtype_suffix

        if not hasattr(self, f_suffixed):
            raise NotImplementedError(f"{self.__class__.__name__}: `{self.name}` is not supported for now. "
                                      "If you need it, please open an issue on the Leaspy repository on Gitlab.")

        return getattr(self, f_suffixed)(*args, **kws)

    return f_sub



import json

from leaspy import __version__
from leaspy.exceptions import LeaspyModelInputError
from leaspy.utils.typing import KwargsType, Union


class ModelSettings:
    """
    Used in :meth:`.Leaspy.load` to create a :class:`.Leaspy` class object from a `json` file.

    Parameters
    ----------
    path_to_model_settings_or_dict : dict or str
        * If a str: path to a json file containing model settings
        * If a dict: content of model settings

    Raises
    ------
    :exc:`.LeaspyModelInputError`
    """
    def __init__(self, path_to_model_settings_or_dict: Union[str, dict]):

        if isinstance(path_to_model_settings_or_dict, dict):
            settings = path_to_model_settings_or_dict
        elif isinstance(path_to_model_settings_or_dict, str):
            with open(path_to_model_settings_or_dict) as fp:
                settings = json.load(fp)
        else:
            raise LeaspyModelInputError(f"Bad type for model settings: should be a dict or a path as a string, not {type(path_to_model_settings_or_dict)}")

        ModelSettings._check_settings(settings)
        self._get_name(settings)
        self._get_parameters(settings)
        self._get_hyperparameters(settings)

    @staticmethod
    def _check_settings(settings):

        error_tpl = "The '{}' key is missing in the model parameters (JSON file) you are loading."

        for mandatory_key in ['name', 'parameters']:
            if mandatory_key not in settings.keys():
                raise LeaspyModelInputError(error_tpl.format(mandatory_key))

        # check leaspy_version attribute for compatibility purposes
        if 'leaspy_version' not in settings.keys():
            raise LeaspyModelInputError("The model you are trying to load was generated with a leaspy version < 1.1 "
                    f"and is not compatible with your current version of leaspy == {__version__} "
                    "because of a bug in the multivariate model which lead to under-optimal results.\n"
                    "Please consider re-calibrating your model with your current leaspy version.\n"
                    "If you really want to load it as is (at your own risk) please use leaspy == 1.0.*")
        else:
            # we will be able to add some checks here to check/adapt retro/future compatibility of models
            pass

    def _get_name(self, settings):
        self.name: str = settings['name'].lower()

    def _get_parameters(self, settings):
        self.parameters: KwargsType = settings['parameters']

    def _get_hyperparameters(self, settings):
        hyperparameters = {k.lower(): v for k, v in settings.items()
                           if k not in ['name', 'parameters', 'leaspy_version']}

        self.hyperparameters: KwargsType = hyperparameters




from __future__ import annotations
from typing import TYPE_CHECKING
from abc import ABC, abstractmethod
import itertools
import json
import warnings

import torch
import numpy as np

from leaspy import __version__
from leaspy.exceptions import LeaspyModelInputError
from leaspy.utils.typing import KwargsType, FeatureType, Optional, List, Tuple

if TYPE_CHECKING:
    from leaspy.io.data.dataset import Dataset


class GenericModel(ABC):
    """
    Generic model (temporary until :class:`.AbstractModel` is really **abstract**).

    TODO: change naming after AbstractModel was renamed?

    Parameters
    ----------
    name : str
        Name of the model
    **kwargs
        Hyperparameters of the model

    Attributes
    ----------
    name : str
        Name of the model
    is_initialized : bool
        Is the model initialized?
    features : list[str]
        List of model features (None if not initialization)
    dimension : int (read-only)
        Number of features
    parameters : dict
        Contains internal parameters of the model
    """

    # to be changed in sub-classes so to benefit from automatic methods

    # dict of {hyperparam_name: (default_value, type_hint)} instead?
    _hyperparameters: KwargsType = {}
    # top-level "hyperparameters" that are FULLY defined by others hyperparameters
    _properties: Tuple[str, ...] = ('dimension',)

    #_parameters = () # names may be dynamic depending on hyperparameters...
    #_attributes = () # TODO: really pertinent? why not a model parameter? cf. "mixing_matrix"

    def __init__(self, name: str, **kwargs):

        self.name = name
        #self.reset_hyperparameters()
        self.features: List[FeatureType] = None
        self.parameters: KwargsType = {}
        #self.dimension = None
        #self.noise_model = None

        self.is_initialized: bool = False # to be explicitly set as True by subclasses if so

        # Load hyperparameters at init (and set at default values when missing)
        self.load_hyperparameters(kwargs, with_defaults=True)

    """
    # TODO?
    def reset_hyperparameters(self) -> None:
        for hp_name, hp_type_hint in self._hyperparameters.items():
            setattr(self, hp_name, None)
            self.__annotations__[hp_name] = hp_type_hint #Optional[hp_type_hint]
    """

    def validate_compatibility_of_dataset(self, dataset: Dataset):
        """
        Raise if the given dataset is not compatible with the current model.

        Parameters
        ----------
        dataset : :class:`.Dataset`
            The dataset we want to model.

        Raises
        ------
        :exc:`~leaspy.exceptions.LeaspyDataInputError`
            If and only if data is incompatible with model.
        """
        return

    def initialize(self, dataset: Dataset, method: str = None):
        """
        Initialize the model given a dataset and an initialization method.

        After calling this method :attr:`is_initialized` should be True and model should be ready for use.

        Parameters
        ----------
        dataset : :class:`.Dataset`
            The dataset we want to initialize from.
        method : str, optional (default None)
            A custom method to initialize the model
        """
        if self.is_initialized and self.features is not None:
            # we also test that self.features is not None, since for `ConstantModel`:
            # `is_initialized`` is True but as a mock for being personalization-ready, without really being initialized!
            warn_msg = '<!> Re-initializing an already initialized model.'
            if dataset.headers != self.features:
                warn_msg += f' Overwritting previous model features ({self.features}) with new ones ({dataset.headers}).'
            warnings.warn(warn_msg)

        self.validate_compatibility_of_dataset(dataset)
        self.features = dataset.headers
        self.is_initialized = True

    def get_hyperparameters(self, *, with_features = True, with_properties = True, default = None) -> KwargsType:
        """
        Get all model hyperparameters

        Parameters
        ----------
        with_features, with_properties : bool (default True)
            Whether to include `features` and respectively all `_properties` (i.e. _dynamic_ hyperparameters)
            in the returned dictionary
        default : Any
            Default value is something is an hyperparameter is missing (should not!)

        Returns
        -------
        dict { hyperparam_name : str -> hyperparam_value : Any }
        """

        hps_names_iters = []

        # <!> Order of hyperparameters matters

        if with_features:
            hps_names_iters.append(['features'])

        hps_names_iters.append(self._hyperparameters.keys())

        if with_properties:
            hps_names_iters.append(self._properties)

        all_hp_names = itertools.chain(*hps_names_iters)

        return {
            hp_name: getattr(self, hp_name, default)
            for hp_name in all_hp_names
        }

    def hyperparameters_ok(self) -> bool:
        """
        Check all model hyperparameters are ok

        Returns
        -------
        bool
        """

        d_ok = {
            hp_name: hp_val is not None #and check hp_val compatible with hp_type_hint
            #for hp_name, hp_type_hint in self._hyperparameters.items()
            for hp_name, hp_val in self.get_hyperparameters(with_features=True, with_properties=True).items()
        }
        return all(d_ok.values())

    # 'features' (and 'dimension') are really core hyperparameters

    @property
    def dimension(self) -> Optional[int]:
        # read-only <-> number of modelled features
        if self.features is None:
            return None
        else:
            return len(self.features)

    """
    # if we want hyperparameters direct access without storing them in top-level
    def __getattr__(self, key: str):# -> Any:
        # overload so to mock hyperparameters on top-class level

    def __hasattr__(self, key: str) -> bool:
        # overload so to mock hyperparameters on top-class level

    def __setattr__(self, key: str, val) -> None:
        # overload so to mock hyperparameters on top-class level
    """

    def load_parameters(self, parameters, *, list_converter=np.array) -> None:
        """
        Instantiate or update the model's parameters.

        Parameters
        ----------
        parameters : dict
            Contains the model's parameters.
        list_converter : callable
            The function to convert list objects.
        """

        """
        self.parameters = {} # reset completely here
        # TODO: optional reset + warn if overwriting existing?
        # TODO: load model defaults at reset instead?
        for k, v in parameters.items():
            self.parameters[k] = v # unserialize here?
        """

        #<!> shallow copy only
        self.parameters = parameters.copy()

        # convert lists
        for k, v in self.parameters.items():
            if isinstance(v, list):
                self.parameters[k] = list_converter(v)

    def load_hyperparameters(self, hyperparameters: KwargsType, *, with_defaults: bool = False) -> None:
        """
        Load model hyperparameters from a dict

        Parameters
        ----------
        hyperparameters : dict[str, Any]
            Contains the model's hyperparameters
        with_defaults : bool (default False)
            If true, it also resets hyperparameters that are part of the model but not included
            in `hyperparameters` to their default value.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            if inconsistent hyperparameters
        """

        # no total reset of hyperparameters here unlike in load_parameters...

        # TODO change this behavior in ModelSettings? why not sending an empty dict instead of None??
        if hyperparameters is None:
            hyperparameters = {}

        settable_hps = {'features'}.union(self._hyperparameters.keys())

        # unknown hyper parameters
        non_settable_hps = set(hyperparameters.keys()).difference(settable_hps)
        # no Python method to get intersection and difference at once... so it is split
        dynamic_hps = non_settable_hps.intersection(self._properties)
        unknown_hps = non_settable_hps.difference(dynamic_hps)

        if len(unknown_hps) > 0:
            raise LeaspyModelInputError(f'Unknown hyperparameters for `{self.__class__.__qualname__}`: {unknown_hps}')

        # set "static" hyperparameters only
        if with_defaults:
            hyperparameters = {**self._hyperparameters, **hyperparameters}

        for hp_name, hp_val in hyperparameters.items():
            if hp_name in settable_hps:
                setattr(self, hp_name, hp_val) # top-level of object...

        # check that dynamic hyperparameters match if provided...
        # (check this after all "static" hyperparameters being set)
        dynamic_hps_given_value_expected_value = {
            d_hp_name: (hyperparameters[d_hp_name], getattr(self, d_hp_name))
            for d_hp_name in dynamic_hps
        }
        dynamic_hps_given_value_neq_expected_value = {
            d_hp_name: (given_v, expected_v)
            for d_hp_name, (given_v, expected_v) in dynamic_hps_given_value_expected_value.items()
            if given_v != expected_v
        }
        if len(dynamic_hps_given_value_neq_expected_value) != 0:
            raise LeaspyModelInputError(f"Dynamic hyperparameters provided do not correspond to the expected ones:\n"
                                        f"{dynamic_hps_given_value_neq_expected_value}")

    def save(self, path: str, **kwargs):
        """
        Save Leaspy object as json model parameter file.

        Default save method: it can be overwritten in child class but should be generic...

        Parameters
        ----------
        path : str
            Path to store the model's parameters.
        **kwargs
            Keyword arguments for json.dump method.
        """
        model_parameters_save = self.parameters.copy() # <!> shallow copy
        for param_name, param_val in model_parameters_save.items():
            if isinstance(param_val, (torch.Tensor, np.ndarray)):
                model_parameters_save[param_name] = param_val.tolist()

        model_settings = {
            'leaspy_version': __version__,
            'name': self.name,
            **self.get_hyperparameters(with_features=True, with_properties=True),
            'parameters': model_parameters_save
        }

        # Default json.dump kwargs:
        kwargs = {'indent': 2, **kwargs}

        with open(path, 'w') as fp:
            json.dump(model_settings, fp, **kwargs)

    @abstractmethod
    def compute_individual_trajectory(self, timepoints, individual_parameters: dict) -> torch.Tensor:
        """
        Compute scores values at the given time-point(s) given a subject's individual parameters.

        Parameters
        ----------
        timepoints : scalar or array_like[scalar] (list, tuple, :class:`numpy.ndarray`)
            Contains the age(s) of the subject.
        individual_parameters : dict[str, Any]
            Contains the individual parameters.
            Each individual parameter should be a scalar or array_like

        Returns
        -------
        :class:`torch.Tensor`
            Contains the subject's scores computed at the given age(s)
            Shape of tensor is (1, n_tpts, n_features)
        """

    def __str__(self):

        lines = [
            f"=== MODEL {self.name} ===" # header
        ]

        # hyperparameters
        for hp_name, hp_val in self.get_hyperparameters(with_features=True, with_properties=True).items():
            lines.append(f"{hp_name} : {hp_val}")

        # separation between hyperparams & params
        lines.append('-'*len(lines[0]))

        for param_name, param_val in self.parameters.items():
            lines.append(f"{param_name} : {param_val}")

        return "\n".join(lines)



import pandas as pd

from leaspy.io.data.dataframe_data_reader import DataframeDataReader


class CSVDataReader(DataframeDataReader):
    """
    Methods to convert `csv files` to `Leaspy`-compliant data containers.

    Only a wrapper for :func:`pandas.read_csv` and :class:`.DataframeDataReader`.

    Parameters
    ----------
    path : str
        Path to the csv file to load (with its extension)
    pd_read_csv_kws : kwargs
        Keyword arguments passed to :func:`pandas.read_csv`
    **df_reader_kws
        Keyword arguments passed to :class:`.DataframeDataReader`

    Raises
    ------
    :exc:`.LeaspyDataInputError`
    """
    def __init__(self, path: str, *, pd_read_csv_kws: dict = {}, **df_reader_kws):

        # enforce ID to be interpreted as string as default (can be overwritten)
        pd_read_csv_kws = {'dtype': {'ID': str}, **pd_read_csv_kws}

        df = pd.read_csv(path, **pd_read_csv_kws)
        super().__init__(df, **df_reader_kws)



from random import shuffle
from collections.abc import Sequence

import torch
from numpy import ndindex

from .abstract_sampler import AbstractSampler
from leaspy.exceptions import LeaspyInputError
from leaspy.utils.docs import doc_with_super
from leaspy.utils.typing import Union, Tuple


@doc_with_super()
class GibbsSampler(AbstractSampler):
    """
    Gibbs sampler class.

    Parameters
    ----------
    info : dict[str, Any]
        The dictionary describing the random variable to sample.
        It should contains the following entries:
            * name : str
            * shape : tuple[int, ...]
            * type : 'population' or 'individual'
    n_patients : int > 0
        Number of patients (useful for individual variables)
    scale : float > 0 or :class:`torch.FloatTensor` > 0
        An approximate scale for the variable.
        It will be used to scale the initial adaptive std-dev used in sampler.
        An extra factor will be applied on top of this scale (hyperparameters):
            * 1% for population parameters (:attr:`.GibbsSampler.STD_SCALE_FACTOR_POP`)
            * 50% for individual parameters (:attr:`.GibbsSampler.STD_SCALE_FACTOR_IND`)
        Note that if you pass a torch tensor, its shape should be compatible with shape of the variable.
    random_order_dimension : bool (default True)
        This parameter controls whether we randomize the order of indices during the sampling loop.
        (only for population variables, since we perform group sampling for individual variables)
        Article https://proceedings.neurips.cc/paper/2016/hash/e4da3b7fbbce2345d7772b0674a318d5-Abstract.html
        gives a rationale on why we should activate this flag.
    mean_acceptation_rate_target_bounds : tuple[lower_bound: float, upper_bound: float] with 0 < lower_bound < upper_bound < 1
        Bounds on mean acceptation rate.
        Outside this range, the adaptation of the std-dev of sampler is triggered
        so to maintain a target acceptation rate in between these too bounds (e.g: ~30%).
    adaptive_std_factor : float in ]0, 1[
        Factor by which we increase or decrease the std-dev of sampler when we are out of
        the custom bounds for the mean acceptation rate. We decrease it by `1 - factor` if too low,
        and increase it with `1 + factor` if too high.
    **base_sampler_kws
        Keyword arguments passed to `AbstractSampler` init method.
        In particular, you may pass the `acceptation_history_length` hyperparameter.

    Attributes
    ----------
    In addition to the attributes present in :class:`.AbstractSampler`:

    std : torch.FloatTensor
        Adaptative std-dev of variable

    Raises
    ------
    :exc:`.LeaspyInputError`
    """

    # Cf. note on `scale` parameter above (heuristic values)
    STD_SCALE_FACTOR_POP = .01
    STD_SCALE_FACTOR_IND = .5

    def __init__(self, info: dict, n_patients: int, *, scale: Union[float, torch.FloatTensor],
                 random_order_dimension: bool = True,
                 mean_acceptation_rate_target_bounds: Tuple[float, float] = (0.2, 0.4),
                 adaptive_std_factor: float = 0.1,
                 **base_sampler_kws):

        super().__init__(info, n_patients, **base_sampler_kws)

        # Scale of variable should always be positive (component-wise if multidimensional)
        if not isinstance(scale, torch.Tensor):
            scale = torch.tensor(scale)
        scale = scale.float()
        if (scale <= 0).any():
            raise LeaspyInputError(f"Scale of variable '{info['name']}' should be positive, not `{scale}`.")

        if info["type"] == "population":
            # Proposition variance is adapted independently on each dimension of the population variable
            self.std = self.STD_SCALE_FACTOR_POP * scale * torch.ones(self.shape)
        elif info["type"] == "individual":
            # Proposition variance is adapted independently on each patient
            true_shape = (n_patients, *self.shape)
            self.std = self.STD_SCALE_FACTOR_IND * scale * torch.ones(true_shape)
        else:
            raise LeaspyInputError(f"Unknown variable type '{info['type']}'.")

        # Internal counter to trigger adaptation of std based on mean acceptation rate
        self._counter: int = 0

        # Torch distribution: all modifications will be in-place on `self.std`
        # So there will be no need to update this distribution!
        self._distribution = torch.distributions.normal.Normal(loc=0.0, scale=self.std)

        # Parameters of the sampler
        self._random_order_dimension = random_order_dimension

        if not (
            isinstance(mean_acceptation_rate_target_bounds, Sequence)
            and len(mean_acceptation_rate_target_bounds) == 2
            and 0 < mean_acceptation_rate_target_bounds[0] < mean_acceptation_rate_target_bounds[1] < 1
        ):
            raise LeaspyInputError("`mean_acceptation_rate_target_bounds` should be a tuple (lower_bound, upper_bound) "
                                   f"with 0 < lower_bound < upper_bound < 1, not '{mean_acceptation_rate_target_bounds}'")
        self._mean_acceptation_lower_bound_before_adaptation, self._mean_acceptation_upper_bound_before_adaptation = mean_acceptation_rate_target_bounds

        if not (0 < adaptive_std_factor < 1):
            raise LeaspyInputError(f"`adaptive_std_factor` should be a float in ]0, 1[, not '{adaptive_std_factor}'")
        self._adaptive_std_factor = adaptive_std_factor

    def __str__(self):
        mean_acceptation_rate = self.acceptation_history.mean().item()  # mean on all dimensions!
        return f"{self.name} rate : {mean_acceptation_rate:.1%}, std: {self.std.mean():.1e}"

    def _proposal(self, val):
        """
        Proposal value around the current value with sampler standard deviation.

        <!> Not to be used for scalar sampling (in `_sample_population_realizations`)
            since it would be inefficient!

        Parameters
        ----------
        val : torch.FloatTensor

        Returns
        -------
        torch.FloatTensor of shape broadcasted_shape(val.shape, self.std.shape)
            value around `val`
        """
        return val + self._distribution.sample()  # sample_shape=val.shape

    def _update_std(self):
        """
        Update standard deviation of sampler according to current frequency of acceptation.

        Adaptive std is known to improve sampling performances.
        For default parameters: std-dev is increased if frequency of acceptation is > 40%,
        and decreased if < 20%, so as to stay close to 30%.
        """
        self._counter += 1

        if self._counter % self.acceptation_history_length == 0:
            mean_acceptation = self.acceptation_history.mean(dim=0)

            idx_toolow = mean_acceptation < self._mean_acceptation_lower_bound_before_adaptation
            idx_toohigh = mean_acceptation > self._mean_acceptation_upper_bound_before_adaptation

            self.std[idx_toolow] *= (1 - self._adaptive_std_factor)
            self.std[idx_toohigh] *= (1 + self._adaptive_std_factor)

    def _sample_population_realizations(self, data, model, realizations, temperature_inv, **attachment_computation_kws):
        """
        For each dimension (1D or 2D) of the population variable, compute current attachment and regularity.
        Propose a new value for the given dimension of the given population variable,
        and compute new attachment and regularity.
        Do a MH step, keeping if better, or if worse with a probability.

        Parameters
        ----------
        data : :class:`.Dataset`
        model : :class:`~.models.abstract_model.AbstractModel`
        realizations : :class:`~.io.realizations.collection_realization.CollectionRealization`
        temperature_inv : float > 0
        **attachment_computation_kws
            Currently not used for population parameters.

        Returns
        -------
        attachment, regularity_var : `torch.FloatTensor` 0D (scalars)
            The attachment and regularity (only for the current variable) at the end of this sampling step (summed on all individuals).
        """
        realization = realizations[self.name]
        shape_current_variable = realization.shape
        accepted_array = torch.zeros(shape_current_variable)
        iterator_indices = list(ndindex(shape_current_variable))
        if self._random_order_dimension:
            shuffle(iterator_indices)  # shuffle in-place!

        # retrieve the individual parameters from realizations once for all to speed-up computations,
        # since they are fixed during the sampling of this population variable!
        ind_params = model.get_param_from_real(realizations)

        def compute_attachment_regularity():
            # model attributes used are the ones from the MCMC toolbox that we are currently changing!
            attachment = model.compute_individual_attachment_tensorized(data, ind_params, attribute_type='MCMC').sum()
            # regularity is always computed with model.parameters (not "temporary MCMC parameters")
            regularity = model.compute_regularity_realization(realization).sum()
            return attachment, regularity

        previous_attachment = previous_regularity = None

        for idx in iterator_indices:
            # Compute the attachment and regularity
            if previous_attachment is None:
                previous_attachment, previous_regularity = compute_attachment_regularity()

            # Keep previous realizations and sample new ones
            old_val_idx = realization.tensor_realizations[idx].clone()
            # the previous version with `_proposal` was not incorrect but computationally inefficient:
            # because we were sampling on the full shape of `std` whereas we only needed `std[idx]` (scalar)
            new_val_idx = old_val_idx + self.std[idx] * torch.randn(())
            realization.set_tensor_realizations_element(new_val_idx, idx)

            # Update derived model attributes if necessary (orthonormal basis, ...)
            model.update_MCMC_toolbox([self.name], realizations)

            # Compute the attachment and regularity
            new_attachment, new_regularity = compute_attachment_regularity()

            alpha = torch.exp(-((new_regularity - previous_regularity) * temperature_inv +
                                (new_attachment - previous_attachment)))

            accepted = self._metropolis_step(alpha)
            accepted_array[idx] = accepted

            if accepted:
                previous_attachment, previous_regularity = new_attachment, new_regularity
            else:
                # Revert modification of realization at idx and its consequences
                realization.set_tensor_realizations_element(old_val_idx, idx)
                # Update (back) derived model attributes if necessary
                # TODO: Shouldn't we backup the old MCMC toolbox instead to avoid heavy computations?
                # (e.g. orthonormal basis re-computation just for a single change)
                model.update_MCMC_toolbox([self.name], realizations)
                # force re-compute on next iteration:
                # not performed since it is useless, since we rolled back to the starting state!
                # previous_attachment = previous_regularity = None

        self._update_acceptation_rate(accepted_array)
        self._update_std()

        # Return last attachment and regularity_var
        return previous_attachment, previous_regularity

    def _sample_individual_realizations(self, data, model, realizations, temperature_inv, **attachment_computation_kws):
        """
        For each individual variable, compute current patient-batched attachment and regularity.
        Propose a new value for the individual variable,
        and compute new patient-batched attachment and regularity.
        Do a MH step, keeping if better, or if worse with a probability.

        Parameters
        ----------
        data : :class:`.Dataset`
        model : :class:`~.models.abstract_model.AbstractModel`
        realizations : :class:`~.io.realizations.collection_realization.CollectionRealization`
        temperature_inv : float > 0
        **attachment_computation_kws
            Optional keyword arguments for attachment computations.
            As of now, we only use it for individual variables, and only `attribute_type`.
            It is used to know whether to compute attachments from the MCMC toolbox (esp. during fit)
            or to compute it from regular model parameters (esp. during personalization in mean/mode realization)

        Returns
        -------
        attachment, regularity_var : `torch.FloatTensor` 1D (n_individuals,)
            The attachment and regularity (only for the current variable) at the end of this sampling step, per individual.
        """

        # Compute the attachment and regularity for all subjects
        realization = realizations[self.name]

        # the population variables are fixed during this sampling step (since we update an individual parameter), but:
        # - if we are in a calibration: we may have updated them just before and have NOT yet propagated these changes
        #   into the master model parameters, so we SHOULD use the MCMC toolbox for model computations (default)
        # - if we are in a personalization (mode/mean real): we are not updating the population parameters any more
        #   so we should NOT use a MCMC_toolbox (not proper)
        attribute_type = attachment_computation_kws.get('attribute_type', 'MCMC')

        def compute_attachment_regularity():
            # current realizations => individual parameters
            ind_params = model.get_param_from_real(realizations)

            # individual parameters => compare reconstructions vs values (per subject)
            attachment = model.compute_individual_attachment_tensorized(data, ind_params, attribute_type=attribute_type)

            # compute log-likelihood of just the given parameter (tau or xi or sources)
            # (per subject; all dimensions of the individual parameter are summed together)
            # regularity is always computed with model.parameters (not "temporary MCMC parameters")
            regularity = model.compute_regularity_realization(realization)
            regularity = regularity.sum(dim=self.ind_param_dims_but_individual).reshape(data.n_individuals)

            return attachment, regularity

        previous_attachment, previous_regularity = compute_attachment_regularity()

        # Keep previous realizations and sample new ones
        previous_reals = realization.tensor_realizations.clone()
        # Add perturbations to previous observations
        realization.tensor_realizations = self._proposal(realization.tensor_realizations)

        # Compute the attachment and regularity
        new_attachment, new_regularity = compute_attachment_regularity()

        # alpha is per patient and > 0, shape = (n_individuals,)
        # if new is "better" than previous, then alpha > 1 so it will always be accepted in `_group_metropolis_step`
        alpha = torch.exp(-((new_regularity - previous_regularity) * temperature_inv +
                            (new_attachment - previous_attachment)))

        accepted = self._group_metropolis_step(alpha)
        self._update_acceptation_rate(accepted)
        self._update_std()

        # compute attachment & regularity
        attachment = accepted*new_attachment + (1-accepted)*previous_attachment
        regularity = accepted*new_regularity + (1-accepted)*previous_regularity

        # we accept together all dimensions of individual parameter
        accepted = accepted.unsqueeze(-1)  # shape = (n_individuals, 1)
        realization.tensor_realizations = accepted*realization.tensor_realizations + (1-accepted)*previous_reals

        return attachment, regularity



import contextlib

import torch

from leaspy.io.data.dataset import Dataset
from leaspy.models.abstract_model import AbstractModel


class AlgoWithDeviceMixin:
    """
    Mixin class containing common attributes & methods for algorithms with a torch device.

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        The specifications of the algorithm as a :class:`.AlgorithmSettings` instance.

    Attributes
    ----------
    algorithm_device : str
        Valid torch device (only the `type` of the device is used)
    """

    def __init__(self, settings):

        super().__init__(settings)

        self.algorithm_device = settings.device

        self._default_algorithm_device = torch.device('cpu')
        self._default_algorithm_tensor_type = 'torch.FloatTensor'

    @contextlib.contextmanager
    def _device_manager(self, model: AbstractModel, dataset: Dataset):
        """
        Context-manager to handle the "ambient device" (i.e. the device used
        to instantiate tensors and perform computations). The provided model
        and dataset will be moved to the device specified for the execution
        at the beginning of the algorithm and moved back to the original
        ('cpu') device at the end of the algorithm. The default tensor type
        will also be set accordingly.

        Parameters
        ----------
        model : :class:`~.models.abstract_model.AbstractModel`
            The used model.
        dataset : :class:`.Dataset`
            Contains the subjects' observations in torch format to speed up computation.
        """
        algorithm_tensor_type = self._default_algorithm_tensor_type
        if self.algorithm_device != self._default_algorithm_device.type:
            algorithm_device = torch.device(self.algorithm_device)

            model.move_to_device(algorithm_device)
            dataset.move_to_device(algorithm_device)

            algorithm_tensor_type = 'torch.cuda.FloatTensor'

        try:
            yield torch.set_default_tensor_type(algorithm_tensor_type)
        finally:
            if self.algorithm_device != self._default_algorithm_device.type:
                model.move_to_device(self._default_algorithm_device)
                dataset.move_to_device(self._default_algorithm_device)

            torch.set_default_tensor_type(self._default_algorithm_tensor_type)



from __future__ import annotations
from typing import TYPE_CHECKING, Union
import warnings

import torch

from leaspy.exceptions import LeaspyInputError, LeaspyModelInputError
from leaspy.utils.typing import TypeVar, KwargsType, Tuple, Callable, Optional, DictParamsTorch
from leaspy.models.utils.noise_struct import NoiseStruct, NOISE_STRUCTS

if TYPE_CHECKING:
    from leaspy.io.data.dataset import Dataset
    from leaspy.models.abstract_model import AbstractModel


T = TypeVar('T')
def constant_return_factory(x: T) -> Callable[[], T]:
    """Helper function to return a function returning the input value."""
    def constant_return():
        return x
    return constant_return


class NoiseModel:
    """
    Helper class to define and work with different noise structures in models.

    TODO? It may be of interest to define an abstract noise structure class with
    all methods needed to be transparently integrated in models and algorithms
    and then to create children classes of it (scalar gaussian, diagonal gaussian,
    more complex gaussian noise structures, Bernoulli realization, ...)

    Parameters
    ----------
    noise_struct : NoiseStruct, or str in `VALID_NOISE_STRUCTS`, or None
        The noise structure to build noise model upon.
        If this is a string, noise structure will be searched among pre-defined noise structures:
            * `'bernoulli'`: Bernoulli realization
            * `'gaussian_scalar'`: Gaussian noise with scalar std-dev, to give as `scale` parameter
            * `'gaussian_diagonal'`: Gaussian noise with 1 std-dev per feature (<!> order), to give as `scale` parameter
        If None: no noise at all (in particular, ultimately `sample(r)_around(values)` will just return `values`)
    **noise_kws
        Keyword arguments to fully characterize the noise structure.
        For now, only one parameter is supported, for Gaussian noise structures:
            * `scale` (:class:`torch.FloatTensor`): the std-dev requested for noise.

    Attributes
    ----------
    struct : NoiseStruct
        The noise structure, which contains its metadata / characteristics.
    distributions_kws : dict[str, Any]
        Extra keyword parameters to be passed to `struct.distribution_factory` apart the centering values.

    Raises
    ------
    :exc:`.LeaspyInputError`
        If `noise_struct` is not supported.

    See Also
    --------
    :meth:`NoiseModel.from_model`
    """

    """Valid structures for noise (except None)."""
    VALID_NOISE_STRUCTS = {k for k in NOISE_STRUCTS.keys() if k is not None}

    """For backward-compatibility only."""
    OLD_MAPPING_FROM_LOSS = {
        'MSE': 'gaussian_scalar',
        'MSE_diag_noise': 'gaussian_diagonal',
        'crossentropy': 'bernoulli'
    }

    def __init__(self, noise_struct: Union[NoiseStruct, str, None], **noise_kws):

        if isinstance(noise_struct, NoiseStruct):
            self.struct = noise_struct
        else:
            self.struct = self.get_named_noise_struct(noise_struct)

        noise_kws_keys = set(noise_kws.keys())
        expected_noise_kws = set(self.struct.model_kws_to_dist_kws.values())

        # Validate the noise keyword arguments that were sent
        pbs = []
        missing_parameters = expected_noise_kws.difference(noise_kws_keys)
        unexpected_parameters = noise_kws_keys.difference(expected_noise_kws)
        if missing_parameters:
            pbs.append(f'should have {missing_parameters} parameters')
        if unexpected_parameters:
            pbs.append(f'should NOT have {unexpected_parameters} parameters')
        if pbs:
            raise LeaspyInputError(f"`noise_struct` = '{noise_struct}' {' but '.join(pbs)}.")

        # validate & clean the noise parameters
        self.distributions_kws = self.struct.validate_dist_kws(noise_kws)

    @classmethod
    def get_named_noise_struct(cls, name: Optional[str]) -> NoiseStruct:
        """Helper to get a default noise structure from its name."""
        noise_struct = NOISE_STRUCTS.get(name, None)

        if noise_struct is None:
            raise LeaspyInputError(f"`noise_struct` = '{name}' is not supported. "
                                   f"Please use one among {cls.VALID_NOISE_STRUCTS} or None.")

        return noise_struct

    def check_compat_with_model(self, model: AbstractModel):
        """Raise if `noise_model` is not compatible with `model` (consistency checks)."""
        self.struct.with_contextual_validators(model=model).validate_dist_kws(self.distributions_kws)

    @classmethod
    def from_model(cls, model: AbstractModel, noise_struct: str = 'model', **noise_kws):
        """
        Initialize a noise model as in the `from_name` initialization but with special keywords
        so to easily inherit noise model from the one of an existing model.

        It also automatically performs some consistency checks between model and noise parameters provided.
        As of now, it is mainly useful for simulation algorithm.

        Parameters
        ----------
        model : :class:`~.AbstractModel`, optional
            The model you want to generate noise for.
            Only used when inheriting noise structure or to perform checks on Gaussian diagonal noise.
        noise_struct : str, optional (default 'model')
            Noise structure requested. Multiple options:
                * 'model': use the noise structure from model, as well as the noise parameters from model (if any)
                * 'inherit_struct' (or deprecated 'default'): use the noise structure from model provided
                (but not the actual parameters of noise from model, if any)
                * All other regular noise structures supported (cf. class docstring)
        **noise_kws : any
            Extra parameters for noise (cf. class docstring)
            Not to be used when `noise_struct` = 'model' (default)

        Returns
        -------
        :class:`.NoiseModel`
        """
        get_noise_parameters_from_model = False

        if noise_struct in ['inherit_struct', 'model', 'default']:

            if noise_struct == 'default':
                warnings.warn("`noise_struct` = 'default' is deprecated and will soon be dropped due to ambiguity, "
                              "use 'inherit_struct' instead for same behavior.", FutureWarning)

            if noise_struct == 'model':
                get_noise_parameters_from_model = True

            # define the 'noise_struct' from the model one (_structure_ only, not parameters unless if it was 'model')
            noise_struct = model.noise_model

        # get default, named, noise structure
        noise_struct_obj = cls.get_named_noise_struct(noise_struct)

        # complete validators of noise structure with dynamic ones, that depend on model information
        noise_struct_obj = noise_struct_obj.with_contextual_validators(model=model)

        if get_noise_parameters_from_model:
            # TODO? we could also only use model noise parameters as default values, possibly overwritten by noise_kws
            if noise_kws:
                raise LeaspyInputError("Extra keyword arguments to specify noise should NOT be provided "
                                       "when `noise_struct` = 'model' in NoiseModel.from_model.")

            # use all noise parameters directly from model parameters (if any)
            # for now there is only 'noise_std' available and for gaussian noise only
            # the mapping of parameters is specific to a noise structure to be general
            model_kws_to_dist_kws = noise_struct_obj.model_kws_to_dist_kws
            noise_kws = {
                model_kws_to_dist_kws[model_param]: model_val
                for model_param, model_val in model.parameters.items()
                if model_param in model_kws_to_dist_kws
            }

        # Instantiate noise model normally (with the special keywords having been substituted)
        # In particular compat with model will be checked here thanks to previous dynamic validators
        return cls(noise_struct_obj, **noise_kws)

    @property
    def scale(self) -> Optional[torch.FloatTensor]:
        """A quick short-cut for scale of Gaussian noises (really useful??)."""
        return self.distributions_kws.get('scale', None)

    def rv_around(self, loc: torch.FloatTensor) -> torch.distributions.Distribution:
        """Return the torch distribution centred around values (only if noise is not None)."""
        if self.struct.distribution_factory is None:
            raise LeaspyInputError('Random variable around values is undefined when there is no noise.')

        return self.struct.distribution_factory(loc, **self.distributions_kws)

    def sampler_around(self, loc: torch.FloatTensor) -> Callable[[], torch.FloatTensor]:
        """Return the noise sampling function around input values."""
        if self.struct.distribution_factory is None:
            # No noise: return raw values (no copy)
            return constant_return_factory(loc)
        else:
            return self.rv_around(loc).sample

    def sample_around(self, model_loc_values: torch.FloatTensor) -> torch.FloatTensor:
        """Realization around `model_loc_values` with respect to noise model."""
        # <!> Better to store sampler if multiple calls needed
        return self.sampler_around(model_loc_values)()

    ## HELPER METHODS ##
    @staticmethod
    def rmse_model(model: AbstractModel, dataset: Dataset, individual_params: DictParamsTorch, *,
                   scalar: bool = None, **computation_kwargs) -> torch.FloatTensor:
        """
        Helper function to compute the root mean square error (RMSE) from model reconstructions.

        Parameters
        ----------
        model : :class:`~.AbstractModel`
            Subclass object of `AbstractModel`.
        dataset : :class:`~.Dataset`
            Dataset to compute reconstruction errors from.
        individual_params : DictParamsTorch
            Object containing the computed individual parameters (torch format).
        scalar : bool or None (default)
            Should we compute a scalar RMSE (averaged on all features) or one RMSE per feature (same order)?
            If None, it will fetch `noise_model` from model and choose scalar mode iif 'scalar' in `noise_model` string.
        **computation_kwargs
            Additional kwargs for `model.compute_sum_squared_***_tensorized` method

        Returns
        -------
        :class:`torch.FloatTensor`
            The RMSE tensor (1D) of length 1 if `scalar` else `model.dimension`.
        """

        if scalar is None:
            scalar = 'scalar' in model.noise_model

        if scalar:
            sum_squared = model.compute_sum_squared_tensorized(dataset, individual_params,
                                                               **computation_kwargs).sum(dim=0)  # sum on individuals
            return torch.sqrt(sum_squared / dataset.n_observations)
        else:
            # 1 noise per feature
            sum_squared_per_ft = model.compute_sum_squared_per_ft_tensorized(dataset, individual_params,
                                                                             **computation_kwargs).sum(dim=0)   # sum on individuals
            return torch.sqrt(sum_squared_per_ft / dataset.n_observations_per_ft.float())

    @classmethod
    def _extract_noise_model_from_old_loss_for_backward_compatibility(cls, loss: Optional[str]) -> Optional[str]:
        """Only for backward-compatibility with old loss."""

        if loss is None:
            return None

        noise_struct = cls.OLD_MAPPING_FROM_LOSS.get(loss, None)
        if noise_struct is None:
            raise LeaspyModelInputError(f'Old loss "{loss}" is not known and not supported anymore. '
                                        f'It should have been in {set(cls.OLD_MAPPING_FROM_LOSS.keys())}.')

        warnings.warn('`loss` hyperparameter of model is deprecated, it should be named `noise_model` from now on. '
                      f'Please replace parameter with: `noise_model` = "{noise_struct}".', FutureWarning)

        return noise_struct

    @classmethod
    def set_noise_model_from_hyperparameters(cls, model, hyperparams: KwargsType) -> Tuple[str, ...]:
        """
        Set `noise_model` of a model from hyperparameters.

        Parameters
        ----------
        model : AbstractModel
            Where to set noise model (in-place)
        hyperparams : dict[str, Any]
            where to look for noise model

        Returns
        -------
        tuple[str]
            Additional recognized hyperparameters for models.
        """

        # BACKWARD-COMPAT
        if 'loss' in hyperparams.keys():
            model.noise_model = cls._extract_noise_model_from_old_loss_for_backward_compatibility(hyperparams['loss'])
        # END

        if 'noise_model' in hyperparams.keys():
            if hyperparams['noise_model'] not in cls.VALID_NOISE_STRUCTS:
                raise LeaspyModelInputError(f'`noise_model` should be in {cls.VALID_NOISE_STRUCTS}, '
                                            f'not "{hyperparams["noise_model"]}".')
            model.noise_model = hyperparams['noise_model']

        return ('loss', 'noise_model')



from abc import abstractmethod
import json
import math
import warnings

import torch

from leaspy import __version__

from leaspy.models.abstract_model import AbstractModel
from leaspy.models.utils.attributes import AttributesFactory
from leaspy.models.utils.attributes.abstract_manifold_model_attributes import AbstractManifoldModelAttributes
from leaspy.models.utils.initialization.model_initialization import initialize_parameters
from leaspy.models.utils.noise_model import NoiseModel

from leaspy.utils.typing import KwargsType, List, Optional
from leaspy.utils.docs import doc_with_super
from leaspy.exceptions import LeaspyModelInputError


@doc_with_super()
class AbstractMultivariateModel(AbstractModel):
    """
    Contains the common attributes & methods of the multivariate models.

    Parameters
    ----------
    name : str
        Name of the model
    **kwargs
        Hyperparameters for the model

    Raises
    ------
    :exc:`.LeaspyModelInputError`
        if inconsistent hyperparameters
    """
    def __init__(self, name: str, **kwargs):

        super().__init__(name)

        self.source_dimension: int = None
        self.noise_model = 'gaussian_diagonal'

        self.parameters = {
            "g": None,
            "betas": None,
            "tau_mean": None, "tau_std": None,
            "xi_mean": None, "xi_std": None,
            "sources_mean": None, "sources_std": None,
            "noise_std": None
        }
        self.bayesian_priors = None
        self.attributes: AbstractManifoldModelAttributes = None

        # MCMC related "parameters"
        self.MCMC_toolbox = {
            'attributes': None,
            'priors': {
                # for logistic: "p0" = 1 / (1+exp(g)) i.e. exp(g) = 1/p0 - 1
                # for linear: "p0" = g
                'g_std': None,
                'betas_std': None
            }
        }

        # Load hyperparameters at end to overwrite default for new hyperparameters
        self.load_hyperparameters(kwargs)

    """
    def smart_initialization_realizations(self, data, realizations):
        # TODO : Qui a fait ça? A quoi ça sert?
        # means_time = torch.tensor([torch.mean(data.get_times_patient(i)) for
        # i in range(data.n_individuals)]).reshape(realizations['tau'].tensor_realizations.shape)
        # realizations['tau'].tensor_realizations = means_time
        return realizations
    """

    def initialize(self, dataset, method: str = 'default'):

        if dataset.dimension < 2:
            raise LeaspyModelInputError("A multivariate model should have at least 2 features but your dataset "
                                        f"only contains {dataset.dimension} features ({dataset.headers}).")

        self.dimension = dataset.dimension
        self.features = dataset.headers

        if self.source_dimension is None:
            self.source_dimension = int(math.sqrt(dataset.dimension))
            warnings.warn('You did not provide `source_dimension` hyperparameter for multivariate model, '
                          f'setting it to ⌊√dimension⌋ = {self.source_dimension}.')

        elif not (isinstance(self.source_dimension, int) and 0 <= self.source_dimension < self.dimension):
            raise LeaspyModelInputError(f"Sources dimension should be an integer in [0, dimension - 1[ "
                                        f"but you provided `source_dimension` = {self.source_dimension} whereas `dimension` = {self.dimension}")

        self.parameters = initialize_parameters(self, dataset, method)

        self.attributes = AttributesFactory.attributes(self.name, self.dimension, self.source_dimension)

        # Postpone the computation of attributes when really needed!
        #self.attributes.update(['all'], self.parameters)

        self.is_initialized = True

    @abstractmethod
    def initialize_MCMC_toolbox(self) -> None:
        """
        Initialize Monte-Carlo Markov-Chain toolbox for calibration of model
        """
        # TODO to move in a "MCMC-model interface"

    @abstractmethod
    def update_MCMC_toolbox(self, name_of_the_variables_that_have_been_changed: List[str], realizations) -> None:
        """
        Update the MCMC toolbox with a collection of realizations of model population parameters.

        Parameters
        ----------
        name_of_the_variables_that_have_been_changed : container[str] (list, tuple, ...)
            Names of the population parameters to update in MCMC toolbox
        realizations : :class:`.CollectionRealization`
            All the realizations to update MCMC toolbox with
        """
        # TODO to move in a "MCMC-model interface"

    def load_hyperparameters(self, hyperparameters: KwargsType):

        expected_hyperparameters = ('features', 'dimension', 'source_dimension')

        if 'features' in hyperparameters.keys():
            self.features = hyperparameters['features']

        if 'dimension' in hyperparameters.keys():
            if self.features and hyperparameters['dimension'] != len(self.features):
                raise LeaspyModelInputError(f"Dimension provided ({hyperparameters['dimension']}) does not match features ({len(self.features)})")
            self.dimension = hyperparameters['dimension']

        if 'source_dimension' in hyperparameters.keys():
            if not (
                isinstance(hyperparameters['source_dimension'], int)
                and (hyperparameters['source_dimension'] >= 0)
                and (not self.dimension or hyperparameters['source_dimension'] <= self.dimension - 1)
            ):
                raise LeaspyModelInputError(f"Source dimension should be an integer in [0, dimension - 1], not {hyperparameters['source_dimension']}")
            self.source_dimension = hyperparameters['source_dimension']

        # load new `noise_model` directly in-place & add the recognized hyperparameters to known tuple
        expected_hyperparameters += NoiseModel.set_noise_model_from_hyperparameters(self, hyperparameters)

        self._raise_if_unknown_hyperparameters(expected_hyperparameters, hyperparameters)

    def save(self, path: str, with_mixing_matrix: bool = True, **kwargs):
        """
        Save Leaspy object as json model parameter file.

        Parameters
        ----------
        path : str
            Path to store the model's parameters.
        with_mixing_matrix : bool (default True)
            Save the mixing matrix in the exported file in its 'parameters' section.
            <!> It is not a real parameter and its value will be overwritten at model loading
            (orthonormal basis is recomputed from other "true" parameters and mixing matrix
            is then deduced from this orthonormal basis and the betas)!
            It was integrated historically because it is used for convenience in browser webtool and only there...
        **kwargs
            Keyword arguments for json.dump method.
            Default to: dict(indent=2)
        """
        model_parameters_save = self.parameters.copy()

        if with_mixing_matrix:
            model_parameters_save['mixing_matrix'] = self.attributes.mixing_matrix

        for key, value in model_parameters_save.items():
            if isinstance(value, torch.Tensor):
                model_parameters_save[key] = value.tolist()

        model_settings = {
            'leaspy_version': __version__,
            'name': self.name,
            'features': self.features,
            'dimension': self.dimension,
            'source_dimension': self.source_dimension,
            'noise_model': self.noise_model,
            'parameters': model_parameters_save
        }

        # Default json.dump kwargs:
        kwargs = {'indent': 2, **kwargs}

        with open(path, 'w') as fp:
            json.dump(model_settings, fp, **kwargs)

    @abstractmethod
    def compute_individual_tensorized(self, timepoints, individual_parameters, *, attribute_type=None) -> torch.FloatTensor:
        pass

    def compute_mean_traj(self, timepoints, *, attribute_type: Optional[str] = None):
        """
        Compute trajectory of the model with individual parameters being the group-average ones.

        TODO check dimensions of io?

        Parameters
        ----------
        timepoints : :class:`torch.Tensor` [1, n_timepoints]
        attribute_type : 'MCMC' or None

        Returns
        -------
        :class:`torch.Tensor` [1, n_timepoints, dimension]
            The group-average values at given timepoints
        """
        individual_parameters = {
            'xi': torch.tensor([self.parameters['xi_mean']]),
            'tau': torch.tensor([self.parameters['tau_mean']]),
            'sources': torch.zeros(self.source_dimension)
        }

        return self.compute_individual_tensorized(timepoints, individual_parameters, attribute_type=attribute_type)

    def _get_attributes(self, attribute_type: Optional[str]):
        if attribute_type is None:
            return self.attributes.get_attributes()
        elif attribute_type == 'MCMC':
            return self.MCMC_toolbox['attributes'].get_attributes()
        else:
            raise LeaspyModelInputError(f"The specified attribute type does not exist: {attribute_type}. "
                                        "Should be None or 'MCMC'.")



from bisect import bisect

from leaspy.exceptions import LeaspyDataInputError
from leaspy.utils.typing import IDType, DictParams, KwargsType, List, Iterable


class IndividualData:
    """
    Data container for individual parameters, used to construct a `Data` container.

    Parameters
    ----------
    idx : str
        The identifier of subject.

    Raises
    ------
    :exc:`.LeaspyDataInputError`
    """

    def __init__(self, idx: IDType):
        self.idx = idx
        self.timepoints: List[float] = None
        self.observations: List[Iterable[float]] = None
        self.individual_parameters: DictParams = {}
        self.cofactors: KwargsType = {}

    def add_observation(self, timepoint, observation):
        if self.timepoints is None:
            self.timepoints = []
            self.observations = []

        if timepoint in self.timepoints:
            raise LeaspyDataInputError(
                f'You are trying to overwrite the observation at time {timepoint} of the subject {self.idx}')

        index = bisect(self.timepoints, timepoint)
        self.timepoints.insert(index, timepoint)
        self.observations.insert(index, observation)

    def add_observations(self, timepoints, observations):

        for i, timepoint in enumerate(timepoints):
            self.add_observation(timepoint, observations[i])

    def add_individual_parameters(self, name, value):
        self.individual_parameters[name] = value

    def add_cofactors(self, d: dict):
        for k, v in d.items():
            if k in self.cofactors.keys() and v != self.cofactors[k]:
                raise LeaspyDataInputError(f"The cofactor {k} is already present for patient {self.idx}")
            self.cofactors[k] = v

    def add_cofactor(self, name, value):
        self.cofactors[name] = value



from abc import ABC, abstractmethod

import torch

from leaspy.exceptions import LeaspyModelInputError
from leaspy.utils.typing import KwargsType, Tuple, Optional

from leaspy.io.data.dataset import Dataset
from leaspy.models.abstract_model import AbstractModel
from leaspy.io.realizations.collection_realization import CollectionRealization


class AbstractSampler(ABC):
    """
    Abstract sampler class.

    Parameters
    ----------
    info : dict[str, Any]
        The dictionary describing the random variable to sample.
        It should contains the following entries:
            * name : str
            * shape : tuple[int, ...]
            * type : 'population' or 'individual'
    n_patients : int > 0
        Number of patients (useful for individual variables)
    acceptation_history_length : int > 0 (default 25)
        Deepness (= number of iterations) of the history kept for computing the mean acceptation rate.
        (It is the same for population or individual variables.)

    Attributes
    ----------
    name : str
        Name of variable
    shape : tuple
        Shape of variable
    acceptation_history_length : int
        Deepness (= number of iterations) of the history kept for computing the mean acceptation rate.
        (It is the same for population or individual variables.)
    ind_param_dims_but_individual : tuple[int, ...], optional (only for individual variable)
        The dimension(s) to aggregate when computing regularity of individual parameters
        For now there's only one extra dimension whether it's tau, xi or sources
        but in the future it could be extended. We do not sum first dimension (=0) which
        will always be the dimension reserved for individuals.
    acceptation_history : :class:`torch.Tensor`
        History of binary acceptations to compute mean acceptation rate for the sampler in MCMC-SAEM algorithm.
        It keeps the history of the last `acceptation_history_length` steps.

    Raises
    ------
    :exc:`.LeaspyModelInputError`
    """

    def __init__(self, info: KwargsType, n_patients: int, *, acceptation_history_length: int = 25):

        self.name: str = info["name"]
        self.shape: Tuple[int, ...] = info["shape"]
        self.acceptation_history_length = acceptation_history_length

        self.ind_param_dims_but_individual: Optional[Tuple[int, ...]] = None

        if info["type"] == "population":
            self.type = 'pop'
            # Initialize the acceptation history
            if len(self.shape) not in {1, 2}:
                # convention: shape of pop variable is 1D or 2D
                raise LeaspyModelInputError("Dimension of population variable should be 1 or 2")
            else:
                full_shape = (self.acceptation_history_length, *self.shape)
                self.acceptation_history = torch.zeros(full_shape)

        elif info["type"] == "individual":
            self.type = 'ind'
            # Initialize the acceptation history
            if len(self.shape) != 1:
                raise LeaspyModelInputError("Dimension of individual variable should be 1")
            # <!> We do not take into account the dimensionality of individual parameter for acceptation rate
            full_shape = (self.acceptation_history_length, n_patients)
            self.acceptation_history = torch.zeros(full_shape)

            # The dimension(s) to sum when computing regularity of individual parameters
            # For now there's only one extra dimension whether it's tau, xi or sources
            # but in the future it could be extended. We never sum dimension 0 which
            # will always be the individual dimension.
            self.ind_param_dims_but_individual = tuple(range(1, 1 + len(self.shape)))  # for now it boils down to (1,)
        else:
            raise LeaspyModelInputError(f"Unknown variable type '{info['type']}': nor 'population' nor 'individual'.")

    def sample(self, dataset: Dataset, model: AbstractModel, realizations: CollectionRealization, temperature_inv: float, **attachment_computation_kws) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        """
        Sample new realization (either population or individual) for a given realization state, dataset, model and temperature

        <!> Modifies in-place the realizations object,
        <!> as well as the model through its `update_MCMC_toolbox` for population variables.

        Parameters
        ----------
        dataset : :class:`.Dataset`
            Dataset class object build with leaspy class object Data, model & algo
        model : :class:`.AbstractModel`
            Model for loss computations and updates
        realizations : :class:`~.io.realizations.collection_realization.CollectionRealization`
            Contain the current state & information of all the variables of interest
        temperature_inv : float > 0
            Inverse of the temperature used in tempered MCMC-SAEM
        **attachment_computation_kws
            Optional keyword arguments for attachment computations.
            As of now, we only use it for individual variables, and only `attribute_type`.
            It is used to know whether to compute attachments from the MCMC toolbox (esp. during fit)
            or to compute it from regular model parameters (esp. during personalization in mean/mode realization)

        Returns
        -------
        attachment, regularity_var : `torch.FloatTensor` 0D (population variable) or 1D (individual variable, with length `n_individuals`)
            The attachment and regularity (only for the current variable) at the end of this sampling step
            (globally or per individual, depending on variable type).
        """
        if self.type == 'pop':
            return self._sample_population_realizations(dataset, model, realizations, temperature_inv, **attachment_computation_kws)
        else:
            return self._sample_individual_realizations(dataset, model, realizations, temperature_inv, **attachment_computation_kws)

    @abstractmethod
    def _sample_population_realizations(self, data, model, realizations, temperature_inv, **attachment_computation_kws) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        """Sample population variables"""

    @abstractmethod
    def _sample_individual_realizations(self, data, model, realizations, temperature_inv, **attachment_computation_kws) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        """Sample individual variables"""

    def _group_metropolis_step(self, alpha: torch.FloatTensor) -> torch.FloatTensor:
        """
        Compute the acceptance decision (0. for False & 1. for True).

        Parameters
        ----------
        alpha : :class:`torch.FloatTensor` > 0

        Returns
        -------
        accepted : :class:`torch.FloatTensor`, same shape as `alpha`
            Acceptance decision (0. or 1.).
        """
        accepted = torch.rand(alpha.size()) < alpha
        return accepted.float()

    def _metropolis_step(self, alpha: float) -> bool:
        """
        Compute the Metropolis acceptance decision.

        If better (alpha>=1): accept
        If worse (alpha<1): accept with probability alpha

        <!> This function is critical for the reproducibility between machines.
        Different architectures might lead to different rounding errors on alpha
        (e.g: 1. - 1e-6 vs 1. + 1e-6). If we were to draw only for alpha < 1 (and not when alpha >= 1),
        then it would cause the internal seed of pytorch to change or not depending on the case
        which would lead to very different results afterwards (all the random numbers would be affected).

        Parameters
        ----------
        alpha : float > 0

        Returns
        -------
        bool
            acceptance decision (False or True)
        """
        # Sample a realization from uniform law
        # Choose to keep iff realization is < alpha (probability alpha)
        # <!> Always draw a number even if it seems "useless" (cf. docstring warning)
        return torch.rand(1).item() < alpha

    def _update_acceptation_rate(self, accepted: torch.FloatTensor):
        """
        Update acceptation rate from history of boolean accepted values for each dimension of each variable
        (except for multivariate individual parameters)

        Parameters
        ----------
        accepted : :class:`torch.FloatTensor` (0. or 1.)

        Raises
        ------
        :exc:`.LeaspyModelInputError`
        """

        # Concatenate the new acceptation result at end of new one (forgetting the oldest acceptation rate)
        old_acceptation_history = self.acceptation_history[1:]

        if self.type == "pop":
            self.acceptation_history = torch.cat([old_acceptation_history, accepted.unsqueeze(0)])
        elif self.type == "ind":
            self.acceptation_history = torch.cat([old_acceptation_history, accepted.unsqueeze(0)])
        else:
            raise LeaspyModelInputError(f"Unknown variable type '{self.type}': nor 'pop' nor 'ind'.")



from __future__ import annotations
from typing import TYPE_CHECKING
import warnings

import numpy as np

from leaspy.algo.abstract_algo import AbstractAlgo
from leaspy.io.outputs.individual_parameters import IndividualParameters
from leaspy.exceptions import LeaspyAlgoInputError
from leaspy.utils.typing import FeatureType, List

if TYPE_CHECKING:
    from leaspy.models.constant_model import ConstantModel
    from leaspy.io.data.dataset import Dataset


class ConstantPredictionAlgorithm(AbstractAlgo):
    r"""
    ConstantPredictionAlgorithm is the algorithm that outputs a constant prediction

    It is associated to :class:`~.models.constant_model.ConstantModel`

    TODO: it should be a child of `AbstractPersonalizeAlgorithm` (refactoring needed)

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        The settings of constant prediction algorithm.
        It may define `prediction_type` (str):
            * ``'last'``: last value seen during calibration (even if NaN) [default],
            * ``'last_known'``: last non NaN value seen during calibration*§,
            * ``'max'``: maximum (=worst) value seen during calibration*§,
            * ``'mean'``: average of values seen during calibration§.

        | \\* <!> depending on features, the `last_known` / `max` value may correspond to different visits.
        | § <!> for a given feature, value will be NaN if and only if all values for this feature were NaN.

    Raises
    ------
    :exc:`.LeaspyAlgoInputError`
        If any invalid setting for the algorithm
    """

    name = 'constant_prediction'
    family = 'personalize'
    deterministic = True

    _prediction_types = {'last', 'last_known', 'max', 'mean'}

    def __init__(self, settings):

        super().__init__(settings)

        if settings.parameters['prediction_type'] not in self._prediction_types:
            raise LeaspyAlgoInputError(f'The `prediction_type` of the constant prediction should be in {self._prediction_types}')

        self.prediction_type = settings.parameters['prediction_type']

    def run_impl(self, model: ConstantModel, dataset: Dataset):
        """
        Main method, refer to abstract definition in :meth:`~.algo.personalize.abstract_personalize_algo.AbstractPersonalizeAlgo.run`.

        Parameters
        ----------
        model : :class:`~.models.constant_model.ConstantModel`
            A subclass object of leaspy `ConstantModel`.
        dataset : :class:`.Dataset`
            Dataset object build with leaspy class objects Data, algo & model

        Returns
        -------
        individual_parameters : :class:`.IndividualParameters`
            Contains individual parameters.
        noise_std : float
            TODO: always 0 for now
        """

        # always overwrite model features (no fit process)
        # TODO? we could fit the model before, only to recover model features, and then check at personalize that is the same (as in others personalize algos...)

        # Always overwrite model features (no fit for constant model...)
        model.initialize(dataset)

        ip = IndividualParameters()
        for it in range(dataset.n_individuals):
            idx = dataset.indices[it]
            times = dataset.get_times_patient(it)
            values = dataset.get_values_patient(it).numpy()
            ind_ip = self._get_individual_last_values(times, values, fts=model.features)

            ip.add_individual_parameters(str(idx), ind_ip)

        return ip, None # TODO? evaluate rmse?

    def _get_individual_last_values(self, times, values, *, fts: List[FeatureType]):
        """
        Parameters
        ----------
        times : :class:`numpy.ndarray` [float]
            shape (n_visits,)

        values : :class:`numpy.ndarray` [float]
            shape (n_visits, n_features)

        fts : list[str]
            Feature names

        Returns
        -------
        dict[ft_name: str, constant_value_to_be_padded]
        """

        # Return the maximum value
        if self.prediction_type == 'max':
            fts_values = np.nanmax(values, axis=0)

        # Return the mean value
        elif self.prediction_type == 'mean':
            fts_values = np.nanmean(values, axis=0)

        # Return the last or last-known
        else:

            # sort by more recent visits first (= patient is older)
            sorted_indices = sorted(range(len(times)), key=times.__getitem__, reverse=True)

            # Sometimes, last value can be a NaN. If this behavior is intended, then return it anyway
            if self.prediction_type == 'last':
                fts_values = values[sorted_indices[0]]

            else: # == 'last_known'

                values_sorted_desc = values[sorted_indices]

                # get first index of values being non nan, with visits ordered by more recent
                last_non_nan_ix_per_ft = (~np.isnan(values_sorted_desc)).argmax(axis=0)
                # 1 feature value will be nan iff feature was nan at all visits
                fts_values = values_sorted_desc[last_non_nan_ix_per_ft, range(values.shape[1])]

        # return a dict with parameters names being features names
        return dict(zip(fts, fts_values))


    def set_output_manager(self, output_settings):
        """
        Not implemented.
        """
        if output_settings is not None:
            warnings.warn('Settings logs in constant prediction algorithm is not supported.')
        return



from __future__ import annotations
from typing import TYPE_CHECKING
import warnings

import statsmodels.api as sm
from statsmodels.regression.mixed_linear_model import MixedLM, MixedLMParams
import torch
import numpy as np

from leaspy.algo.abstract_algo import AbstractAlgo
from leaspy.models.lme_model import LMEModel
from leaspy.exceptions import LeaspyDataInputError

if TYPE_CHECKING:
    from leaspy.io.data.dataset import Dataset


class LMEFitAlgorithm(AbstractAlgo): # AbstractFitAlgo not so generic (EM)
    """
    Calibration algorithm associated to :class:`~.models.lme_model.LMEModel`

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        * with_random_slope_age : bool
            If False: only varying intercepts
            If True: random intercept & random slope w.r.t ages

            .. deprecated:: 1.2

            You should rather define this directly as an hyperparameter of LME model.
        * force_independent_random_effects : bool
            Force independence of random intercept & random slope
        * other keyword arguments passed to :meth:`statsmodels.regression.mixed_linear_model.MixedLM.fit`

    See Also
    --------
    :class:`statsmodels.regression.mixed_linear_model.MixedLM`
    """

    name = 'lme_fit'
    family = 'fit'

    def __init__(self, settings):

        super().__init__(settings)

        params = settings.parameters.copy()

        # Get model hyperparameters from fit algorithm parameters
        # deprecated only for backward-compat (to be removed soon...),
        self._model_hyperparams_to_set = {
            hp_name: params.pop(hp_name)
            for hp_name in LMEModel._hyperparameters.keys()
            if hp_name in params
        }

        # Algorithm true parameters
        self.force_independent_random_effects = params.pop('force_independent_random_effects')

        # Remaining parameters are parameters of statsmodels `fit` method
        self.sm_fit_parameters = params # popped from other params

    def run_impl(self, model: LMEModel, dataset: Dataset):
        """
        Main method, refer to abstract definition in :meth:`~.algo.fit.abstract_fit_algo.AbstractFitAlgo.run`.

        TODO fix proper inheritance

        Parameters
        ----------
        model : :class:`~.LMEModel`
            A subclass object of leaspy `LMEModel`.
        dataset : :class:`.Dataset`
            Dataset object build with leaspy class objects Data, algo & model

        Returns
        -------
        2-tuple:
            * None
            * noise scale (std-dev), scalar
        """

        # DEPRECATED - TO BE REMOVED: Store some algo "hyperparameters" for the model
        model_hps_in_algo_settings = {
            hp: v for hp, v in self._model_hyperparams_to_set.items()
            if hp in model._hyperparameters.keys() and v is not None
        }
        if len(model_hps_in_algo_settings) != 0:
            warnings.warn(f'You should define {model_hps_in_algo_settings} directly as hyperparameters of LME model. '
                          'The current behaviour will soon be dropped.', FutureWarning)
            model.load_hyperparameters(model_hps_in_algo_settings)
        # END DEPRECATED

        # get data
        ages = self._get_reformated(dataset, 'timepoints')
        ages_mean, ages_std = np.mean(ages).item(), np.std(ages).item()
        ages_norm = (ages - ages_mean) / ages_std

        y = self._get_reformated(dataset, 'values')
        subjects_with_repeat = self._get_reformated_subjects(dataset)

        # model
        X = sm.add_constant(ages_norm, prepend=True, has_constant='add')

        if model.with_random_slope_age:
            exog_re = X

            if self.force_independent_random_effects:
                free = MixedLMParams.from_components(
                    fe_params=np.ones(2),
                    cov_re=np.eye(2)
                )
                self.sm_fit_parameters['free'] = free
                methods_not_compat_with_free = {'powell','nm'}.intersection(self.sm_fit_parameters['method']) # cf. statsmodels doc
                if len(methods_not_compat_with_free) > 0:
                    warnings.warn("<!> Methods {'powell','nm'} are not compatible with `force_independent_random_effects`")
        else:
            exog_re = None # random_intercept only

        lme = MixedLM(y, X, subjects_with_repeat, exog_re, missing='raise')
        fitted_lme = lme.fit(**self.sm_fit_parameters)

        try:
            cov_re_unscaled_inv = np.linalg.inv(fitted_lme.cov_re_unscaled)
        except np.linalg.LinAlgError:
            raise LeaspyDataInputError("Cannot predict random effects from "
                                       "singular covariance structure.")

        parameters = {
            "ages_mean": ages_mean,
            "ages_std": ages_std,
            "fe_params": fitted_lme.fe_params,
            "cov_re": fitted_lme.cov_re,
            "cov_re_unscaled_inv": cov_re_unscaled_inv, # really useful for personalization
            "noise_std": fitted_lme.scale ** .5, # statsmodels scale is variance
            "bse_fe": fitted_lme.bse_fe,
            "bse_re": fitted_lme.bse_re
        }

        # update model parameters
        model.load_parameters(parameters)

        # return `(fitted_lme.resid ** 2).mean() ** .5` instead of scale?
        return None, parameters['noise_std']

    @staticmethod
    def _get_reformated(dataset, elem):
        # reformat ages
        dataset_elem = getattr(dataset, elem)
        # flatten
        flat_elem = torch.flatten(dataset_elem).numpy()
        # remove padding & nans
        final_elem = flat_elem[torch.flatten(dataset.mask > 0)]
        return final_elem

    @staticmethod
    def _get_reformated_subjects(dataset):
        subjects_with_repeat = []
        for ind, subject in enumerate(dataset.indices):
            subjects_with_repeat += [subject]*max(dataset.n_visits_per_individual) #[ind]
        subjects_with_repeat = np.array(subjects_with_repeat)
        # remove padding & nans
        subjects_with_repeat = subjects_with_repeat[torch.flatten(dataset.mask > 0)]
        return subjects_with_repeat

    def set_output_manager(self, output_settings):
        """
        Not implemented.
        """
        if output_settings is not None:
            warnings.warn('Settings logs in lme fit algorithm is not supported.')
        return



from leaspy.algo.fit.abstract_mcmc import AbstractFitMCMC


class TensorMCMCSAEM(AbstractFitMCMC):
    """
    Main algorithm for MCMC-SAEM.

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        MCMC fit algorithm settings

    See Also
    --------
    :class:`.AbstractFitMCMC`
    """

    name = 'mcmc_saem'  # OLD: "MCMC_SAEM (tensor)"



import torch

from leaspy.models.generic_model import GenericModel
from leaspy.exceptions import LeaspyModelInputError

from leaspy.utils.docs import doc_with_super


@doc_with_super()
class ConstantModel(GenericModel):
    r"""
    `ConstantModel` is a benchmark model that predicts constant values (no matter what the patient's ages are).

    These constant values depend on the algorithm setting and the patient's values provided during calibration.
    It could predict:
        * `last`: last value seen during calibration (even if NaN),
        * `last_known`: last non NaN value seen during calibration*§,
        * `max`: maximum (=worst) value seen during calibration*§,
        * `mean`: average of values seen during calibration§.

    | \\* <!> depending on features, the `last_known` / `max` value may correspond to different visits.
    | § <!> for a given feature, value will be NaN if and only if all values for this feature were NaN.

    Parameters
    ----------
    name : str
        The model's name
    **kwargs
        Hyperparameters for the model.
        None supported for now.

    Attributes
    ----------
    name : str
        The model's name
    is_initialized : bool
        Always True (no true initialization needed for constant model)
    features : list[str]
        List of the model features.
        Unlike most models features will be determined at `personalization` only (because it does not needed any `fit`)
    dimension : int
        Number of features (read-only)
    parameters : dict
        Model has no parameters: empty dictionary.
        The `prediction_type` parameter should be defined during `personalization`.
        Example:
            >>> AlgorithmSettings('constant_prediction', prediction_type='last_known')

    See Also
    --------
    :class:`~leaspy.algo.others.constant_prediction_algo.ConstantPredictionAlgorithm`
    """

    def __init__(self, name: str, **kwargs):

        super().__init__(name, **kwargs)

        # no fit algorithm is needed for constant model; every "personalization" will re-initialize model
        # however, we need to mock that model is personalization-ready by setting self.is_initialized (API requirement)
        self.is_initialized = True

    def compute_individual_trajectory(self, timepoints, individual_parameters):

        if self.features is None:
            raise LeaspyModelInputError('The model was not properly initialized.')

        values = [individual_parameters[f] for f in self.features]
        return torch.tensor([[values] * len(timepoints)], dtype=torch.float32)



from __future__ import annotations
from typing import TYPE_CHECKING
from dataclasses import dataclass, field
from functools import reduce
import copy

import torch

from leaspy.exceptions import LeaspyInputError
from leaspy.utils.typing import KwargsType, Tuple, Callable, Optional, Dict, DictParamsTorch

if TYPE_CHECKING:
    from leaspy.models.abstract_model import AbstractModel

# Type aliases
ValidationFunc = Callable[[KwargsType], KwargsType]


@dataclass(frozen=True)
class NoiseStruct:
    """
    Class storing all metadata of a noise structure (read-only).

    This class is not intended to be used directly, it serves as configuration for NoiseModel helper class.

    TODO? really have everything related to noise here, including stuff that is currently hardcoded in models
    (model log-likelihood...)?

    Parameters
    ----------
    distribution_factory : function [torch.Tensor, **kws] -> torch.distributions.Distribution (or None)
        A function taking a :class:`torch.Tensor` of values first, possible keyword arguments
        and returning a noise generator (instance of class :class:`torch.distributions.Distribution`),
        which can sample around these values with respect to noise structure.
    model_kws_to_dist_kws : dict[str, str]
        Mapping from naming of noise parameters in Leaspy model to the related torch distribution parameters.
    dist_kws_validators : tuple[ValidationFunc (kwargs -> kwargs)]
        Tuple of functions that sequentially (FIFO) check (& possibly clean) distribution parameters (input).
        It may raise (LeaspyAlgoInputError) if those are not appropriate for the noise structure.
        Those validators are the ones that we already may define without any need for a context
        (e.g. a 'gaussian_scalar' noise will need the scale to be of dimension 1, always)
    contextual_dist_kws_validators : tuple[**context -> ValidationFunc or None]
        Tuple of functions which are factory of validators functions, based on context parameters.
        Indeed, sometimes we may want to enforce some conditions, but we cannot enforce them without having extra contextual information
        (e.g. the scale of 'gaussian_diagonal' can be of any length in general, but if we already know the model dimension,
         then we want to make sure that the scale parameter will be of the same dimension)
        Note: if a given context is not sufficient to build a validator, factory should return None instead of a ValidationFunc.
        cf. :meth:`NoiseStruct.with_contextual_validators` for more details.

    Attributes
    ----------
    dist_kws_to_model_kws : dict[str, str] (read-only property)
        Mapping from torch distribution parameters to the related noise parameter naming in Leaspy model.

    All the previous parameters are also attributes (dataclass)
    """
    distribution_factory: Optional[Callable[..., torch.distributions.Distribution]] = None
    model_kws_to_dist_kws: Dict[str, str] = field(default_factory=dict)
    dist_kws_validators: Tuple[ValidationFunc, ...] = ()
    contextual_dist_kws_validators: Tuple[Callable[..., Optional[ValidationFunc]], ...] = ()

    @property
    def dist_kws_to_model_kws(self):
        """Shortcut for reciprocal mapping of `model_kws_to_dist_kws`"""
        return {v: k for k, v in self.model_kws_to_dist_kws.items()}

    def validate_dist_kws(self, dist_kws: KwargsType) -> KwargsType:
        """Sequentially compose all validators to validate input."""
        return reduce(
            lambda kws, V: V(kws),
            self.dist_kws_validators,  # sequence of validators (V)
            dist_kws  # initial keywords
        )

    def with_contextual_validators(self, **context_kws):
        """
        Clone the current noise structure but with the additional contextual `dist_kws_validators`.

        Note: the contextual validators will be appended, in FIFO order, to the already existing `dist_kws_validators`
        (so in particular they will be executed after them).

        Parameters
        ----------
        **context_kws
            Any relevant keyword argument which may help to define additional contextual `dist_kws_validators`.

        Returns
        -------
        NoiseStruct
            A cloned version of the current noise structure with relevant extra contextual validators set
            (they are now "static", i.e. regular validators)
        """
        # depending on context, determine which `contextual_dist_kws_validators` are relevant (= not None)
        # and those which are not (= None)
        possible_extra_validators = (ctxt_V(**context_kws) for ctxt_V in self.contextual_dist_kws_validators)

        relevant_extra_dist_kws_validators = tuple(V for V in possible_extra_validators if V is not None)
        # only keep contextual validators that were not relevant at this step (for chaining)
        remaining_contextual_dist_kws_validators = tuple(
            ctxt_V for ctxt_V, V in zip(self.contextual_dist_kws_validators, possible_extra_validators)
            if V is None
        )

        return self.__class__(
            distribution_factory=self.distribution_factory,
            model_kws_to_dist_kws=copy.deepcopy(self.model_kws_to_dist_kws),
            dist_kws_validators=self.dist_kws_validators + relevant_extra_dist_kws_validators,
            contextual_dist_kws_validators=remaining_contextual_dist_kws_validators
        )

# Helpers for validation
def convert_input_to_1D_float_tensors(d: KwargsType) -> DictParamsTorch:
    """Helper function to convert all input values into 1D torch float tensors."""
    return {
        k: (v if isinstance(v, torch.Tensor) else torch.tensor(v)).float().view(-1)
        for k, v in d.items()
    }

def validate_dimension_of_scale_factory(error_tpl: str, expected_dim: int, *,
                                        klass = LeaspyInputError):
    """Helper to produce a validator function that check dimension of scale among parameters."""
    def _validator(d: KwargsType):
        noise_scale = d['scale']  # precondition: is a tensor
        dim_noise_scale = noise_scale.numel()
        if dim_noise_scale != expected_dim:
            raise klass(error_tpl.format(noise_scale=noise_scale, dim_noise_scale=dim_noise_scale))
        return d
    return _validator

check_scale_is_univariate = validate_dimension_of_scale_factory(
    "You have provided a noise `scale` ({noise_scale}) of dimension {dim_noise_scale} "
    "whereas the `noise_struct` = 'gaussian_scalar' you requested requires a "
    "univariate scale (e.g. `scale = 0.1`).",
    expected_dim=1
)

def check_scale_is_compat_with_model_dimension(*, model: AbstractModel, **unused_extra_kws):
    """Check that scale parameter is compatible with model dimension."""
    return validate_dimension_of_scale_factory(
        "You requested a 'gaussian_diagonal' noise. However, the attribute `scale` you gave has "
        f"{{dim_noise_scale}} elements, which mismatches with model dimension of {model.dimension}. "
        f"Please give a list of std-dev for every features {model.features}, in order.",
        expected_dim=model.dimension
    )

def check_scale_is_positive(d: KwargsType):
    """Checks scale of noise is positive (component-wise if not scalar)."""
    noise_scale = d['scale']  # precondition: is a tensor
    if (noise_scale <= 0).any():
        raise LeaspyInputError(f"The noise `scale` parameter should be > 0, which is not the case in {noise_scale}.")
    return d

# Define default noise structures
NOISE_STRUCTS = {

    None: NoiseStruct(),

    'bernoulli': NoiseStruct(
        distribution_factory=torch.distributions.bernoulli.Bernoulli
    ),

    'gaussian_scalar': NoiseStruct(
        distribution_factory=torch.distributions.normal.Normal,
        model_kws_to_dist_kws={'noise_std': 'scale'},
        dist_kws_validators=(convert_input_to_1D_float_tensors, check_scale_is_positive, check_scale_is_univariate)
    ),

    'gaussian_diagonal': NoiseStruct(
        distribution_factory=torch.distributions.normal.Normal,
        model_kws_to_dist_kws={'noise_std': 'scale'},
        dist_kws_validators=(convert_input_to_1D_float_tensors, check_scale_is_positive),
        contextual_dist_kws_validators=(check_scale_is_compat_with_model_dimension,)
    ),
}



"""Define some useful type aliases for static type checks and better input understanding."""

from typing import *

import torch

# Generic dictionary of keyword arguments
KwargsType = Dict[str, Any]

# Type for identifier of individuals
IDType = str

# Type for parameters / variables (mostly in dictionary)
ParamType = str

# Type for feature names
FeatureType = str

DictParams = Dict[ParamType, Any]
DictParamsTorch = Dict[ParamType, torch.FloatTensor]



import warnings

from leaspy.exceptions import LeaspyAlgoInputError


class AlgoWithAnnealingMixin:
    """
    Mixin to use in algorithms needing `temperature_inv`; inherit from this class first.

    Note that this mixin is to be used with a class inheriting from `AbstractAlgo`
    (and in particular that have a `algo_parameters` attribute)

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        The specifications of the algorithm as a :class:`.AlgorithmSettings` instance.

        Please note that you can customize the number of iterations with annealing by setting either:
        * `annealing.n_iter` directly (deprecated but has priority over following setting, not defined by default)
        * `annealing.n_iter_frac`, such that iterations with annealing is a ratio of algorithm `n_iter` (default of 50%)

    Attributes
    ----------
    annealing_on : bool
        Is annealing activated or not?

    temperature : float >= 1
    temperature_inv : float in ]0, 1]
        Temperature and its inverse when using annealing
    """

    def __init__(self, settings):
        super().__init__(settings)

        self.temperature: float = 1.
        self.temperature_inv: float = 1.

        # useful property derived from algo parameters
        self.annealing_on: bool = self.algo_parameters.get('annealing', {}).get('do_annealing', False)
        self._annealing_period: int = None
        self._annealing_temperature_decrement: float = None

        if not self.annealing_on:
            return

        # Dynamic number of iterations for annealing
        annealing_n_iter_frac = self.algo_parameters['annealing']['n_iter_frac']

        if self.algo_parameters['annealing'].get('n_iter', None) is None:
            if annealing_n_iter_frac is None:
                raise ValueError("You should NOT have both `annealing.n_iter_frac` and `annealing.n_iter` None."
                                 "\nPlease set a value for at least one of those settings.")

            self.algo_parameters['annealing']['n_iter'] = int(annealing_n_iter_frac * self.algo_parameters['n_iter'])

        elif annealing_n_iter_frac is not None:
            warnings.warn("`annealing.n_iter` setting is deprecated in favour of `annealing.n_iter_frac` - "
                          "which defines the duration with annealing as a ratio of the total number of iterations."
                          "\nPlease use the new setting to suppress this warning "
                          "or explicitly set `annealing.n_iter_frac=None`."
                          "\nHowever, note that while `annealing.n_iter` is supported "
                          "it will always have priority over `annealing.n_iter_frac`.", FutureWarning)

    def __str__(self):
        out = super().__str__()
        if self.annealing_on:
            out += "\n= Annealing =\n"
            out += f"    temperature : {self.temperature:.1f}"
        return out

    def _initialize_annealing(self):
        """
        Initialize annealing, setting initial temperature and number of iterations.
        """
        if not self.annealing_on:
            return

        self.temperature = self.algo_parameters['annealing']['initial_temperature']
        self.temperature_inv = 1 / self.temperature

        if not (isinstance(self.algo_parameters['annealing']['n_plateau'], int)
                and self.algo_parameters['annealing']['n_plateau'] > 0):
            raise LeaspyAlgoInputError('Your `annealing.n_plateau` should be a positive integer')

        if self.algo_parameters['annealing']['n_plateau'] == 1:
            warnings.warn('You defined `annealing.n_plateau` = 1, so you will stay at initial temperature. '
                          'Consider setting `annealing.n_plateau` >= 2 for a true annealing scheme.')
            return

        self._annealing_period = (
            self.algo_parameters['annealing']['n_iter']
            // (self.algo_parameters['annealing']['n_plateau'] - 1)
        )

        self._annealing_temperature_decrement = (
            (self.algo_parameters['annealing']['initial_temperature'] - 1.)
            / (self.algo_parameters['annealing']['n_plateau'] - 1)
        )
        if self._annealing_temperature_decrement <= 0:
            raise LeaspyAlgoInputError('Your `initial_temperature` should be > 1')

    def _update_temperature(self):
        """
        Update the temperature according to a plateau annealing scheme.
        """
        if not self.annealing_on or self._annealing_period is None:
            return

        if self.current_iteration <= self.algo_parameters['annealing']['n_iter']:
            # If we cross a plateau step
            if self.current_iteration % self._annealing_period == 0:
                # Decrease temperature linearly
                self.temperature -= self._annealing_temperature_decrement
                self.temperature = max(self.temperature, 1)
                self.temperature_inv = 1 / self.temperature



import torch

from leaspy.models.abstract_multivariate_model import AbstractMultivariateModel
from leaspy.models.utils.attributes import AttributesFactory
from leaspy.models.utils.noise_model import NoiseModel

from leaspy.utils.docs import doc_with_super, doc_with_
from leaspy.utils.subtypes import suffixed_method
from leaspy.exceptions import LeaspyModelInputError

# TODO refact? implement a single function
# compute_individual_tensorized(..., with_jacobian: bool) -> returning either model values or model values + jacobians wrt individual parameters
# TODO refact? subclass or other proper code technique to extract model's concrete formulation depending on if linear, logistic, mixed log-lin, ...


@doc_with_super()
class MultivariateModel(AbstractMultivariateModel):
    """
    Manifold model for multiple variables of interest (logistic or linear formulation).

    Parameters
    ----------
    name : str
        Name of the model
    **kwargs
        Hyperparameters of the model

    Raises
    ------
    :exc:`.LeaspyModelInputError`
        * If `name` is not one of allowed sub-type: 'univariate_linear' or 'univariate_logistic'
        * If hyperparameters are inconsistent
    """

    SUBTYPES_SUFFIXES = {
        'linear': '_linear',
        'logistic': '_logistic',
        'mixed_linear-logistic': '_mixed',
    }

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self.parameters["v0"] = None
        self.MCMC_toolbox['priors']['v0_std'] = None  # Value, Coef

        self._subtype_suffix = self._check_subtype()

        # enforce a prior for v0_mean --> legacy / never used in practice
        self._set_v0_prior = False


    def _check_subtype(self):
        if self.name not in self.SUBTYPES_SUFFIXES.keys():
            raise LeaspyModelInputError(f'Multivariate model name should be among these valid sub-types: '
                                        f'{list(self.SUBTYPES_SUFFIXES.keys())}.')

        return self.SUBTYPES_SUFFIXES[self.name]

    def load_parameters(self, parameters):
        # TODO? Move this method in higher level class AbstractMultivariateModel? (<!> Attributes class)
        self.parameters = {}
        for k in parameters.keys():
            if k in ['mixing_matrix']:
                continue
            self.parameters[k] = torch.tensor(parameters[k])

        # derive the model attributes from model parameters upon reloading of model
        self.attributes = AttributesFactory.attributes(self.name, self.dimension, self.source_dimension)
        self.attributes.update(['all'], self.parameters)

    @suffixed_method
    def compute_individual_tensorized(self, timepoints, individual_parameters, *, attribute_type=None):
        pass

    def compute_individual_tensorized_linear(self, timepoints, individual_parameters, *, attribute_type=None):

        # Population parameters
        positions, velocities, mixing_matrix = self._get_attributes(attribute_type)
        xi, tau = individual_parameters['xi'], individual_parameters['tau']
        reparametrized_time = self.time_reparametrization(timepoints, xi, tau)

        # Reshaping
        reparametrized_time = reparametrized_time.unsqueeze(-1)  # for automatic broadcast on n_features (last dim)

        # Model expected value
        model = positions + velocities * reparametrized_time

        if self.source_dimension != 0:
            sources = individual_parameters['sources']
            wi = sources.matmul(mixing_matrix.t())
            model += wi.unsqueeze(-2)

        return model # (n_individuals, n_timepoints, n_features)

    def compute_individual_tensorized_logistic(self, timepoints, individual_parameters, *, attribute_type=None):

        # Population parameters
        g, v0, a_matrix = self._get_attributes(attribute_type)
        g_plus_1 = 1. + g
        b = g_plus_1 * g_plus_1 / g

        # Individual parameters
        xi, tau = individual_parameters['xi'], individual_parameters['tau']
        reparametrized_time = self.time_reparametrization(timepoints, xi, tau)

        # Reshaping
        reparametrized_time = reparametrized_time.unsqueeze(-1) # (n_individuals, n_timepoints, n_features)

        # Model expected value
        t = v0 * reparametrized_time
        if self.source_dimension != 0:
            sources = individual_parameters['sources']
            wi = sources.matmul(a_matrix.t())
            t += wi.unsqueeze(-2) # unsqueeze for (n_timepoints)

        # TODO? more efficient & accurate to compute `torch.exp(-t*b + log_g)` since we directly sample & stored log_g
        model = 1. / (1. + g * torch.exp(-t * b))

        return model # (n_individuals, n_timepoints, n_features)

    @suffixed_method
    def compute_individual_ages_from_biomarker_values_tensorized(self, value: torch.Tensor,
                                                                 individual_parameters: dict, feature: str):
        pass

    def compute_individual_ages_from_biomarker_values_tensorized_logistic(self, value: torch.Tensor,
                                                                          individual_parameters: dict, feature: str):
        if value.dim() != 2:
            raise LeaspyModelInputError(f"The biomarker value should be dim 2, not {value.dim()}!")

        # avoid division by zero:
        value = value.masked_fill((value == 0) | (value == 1), float('nan'))

        # 1/ get attributes
        g, v0, a_matrix = self._get_attributes(None)
        xi, tau = individual_parameters['xi'], individual_parameters['tau']
        if self.source_dimension != 0:
            sources = individual_parameters['sources']
            wi = sources.matmul(a_matrix.t())
        else:
            wi = 0

        # get feature value for g, v0 and wi
        feat_ind = self.features.index(feature)  # all consistency checks were done in API layer
        g = torch.tensor([g[feat_ind]])  # g and v0 were shape: (n_features in the multivariate model)
        v0 = torch.tensor([v0[feat_ind]])
        if self.source_dimension != 0:
            wi = wi[0, feat_ind].item()  # wi was shape (1, n_features)

        # 2/ compute age
        ages = tau + (torch.exp(-xi) / v0) * ((g / (g + 1) ** 2) * torch.log(g/(1 / value - 1)) - wi)
        # assert ages.shape == value.shape

        return ages

    @suffixed_method
    def compute_jacobian_tensorized(self, timepoints, individual_parameters, *, attribute_type=None):
        pass

    def compute_jacobian_tensorized_linear(self, timepoints, individual_parameters, *, attribute_type=None):

        # Population parameters
        _, v0, mixing_matrix = self._get_attributes(attribute_type)

        # Individual parameters
        xi, tau = individual_parameters['xi'], individual_parameters['tau']
        reparametrized_time = self.time_reparametrization(timepoints, xi, tau)

        # Reshaping
        reparametrized_time = reparametrized_time.unsqueeze(-1) # (n_individuals, n_timepoints, n_features)
        alpha = torch.exp(xi).reshape(-1, 1, 1)
        dummy_to_broadcast_n_ind_n_tpts = torch.ones_like(reparametrized_time)

        # Jacobian of model expected value w.r.t. individual parameters
        derivatives = {
            'xi': (v0 * reparametrized_time).unsqueeze(-1), # add a last dimension for len param
            'tau': (v0 * -alpha * dummy_to_broadcast_n_ind_n_tpts).unsqueeze(-1), # same
        }

        if self.source_dimension > 0:
            derivatives['sources'] = mixing_matrix.expand((1,1,-1,-1)) * dummy_to_broadcast_n_ind_n_tpts.unsqueeze(-1)

        # dict[param_name: str, torch.Tensor of shape(n_ind, n_tpts, n_fts, n_dims_param)]
        return derivatives

    def compute_jacobian_tensorized_logistic(self, timepoints, individual_parameters, *, attribute_type=None):
        # TODO: refact highly inefficient (many duplicated code from `compute_individual_tensorized_logistic`)

        # Population parameters
        g, v0, a_matrix = self._get_attributes(attribute_type)
        g_plus_1 = 1. + g
        b = g_plus_1 * g_plus_1 / g

        # Individual parameters
        xi, tau = individual_parameters['xi'], individual_parameters['tau']
        reparametrized_time = self.time_reparametrization(timepoints, xi, tau)

        # Reshaping
        reparametrized_time = reparametrized_time.unsqueeze(-1) # (n_individuals, n_timepoints, n_features)
        alpha = torch.exp(xi).reshape(-1, 1, 1)

        # Model expected value
        t = v0 * reparametrized_time
        if self.source_dimension != 0:
            sources = individual_parameters['sources']
            wi = sources.matmul(a_matrix.t())
            t += wi.unsqueeze(-2)
        model = 1. / (1. + g * torch.exp(-t * b))

        # Jacobian of model expected value w.r.t. individual parameters
        c = model * (1. - model) * b

        derivatives = {
            'xi': (c * v0 * reparametrized_time).unsqueeze(-1),
            'tau': (c * v0 * -alpha).unsqueeze(-1),
        }
        if self.source_dimension > 0:
            derivatives['sources'] = c.unsqueeze(-1) * a_matrix.expand((1,1,-1,-1))

        # dict[param_name: str, torch.Tensor of shape(n_ind, n_tpts, n_fts, n_dims_param)]
        return derivatives

    ##############################
    ### MCMC-related functions ###
    ##############################

    def initialize_MCMC_toolbox(self):
        self.MCMC_toolbox = {
            'priors': {'g_std': 0.01, 'v0_std': 0.01, 'betas_std': 0.01}, # population parameters
            'attributes': AttributesFactory.attributes(self.name, self.dimension, self.source_dimension)
        }

        # Initialize a prior for v0_mean (legacy code / never used in practice)
        if self._set_v0_prior:
            self.MCMC_toolbox['priors']['v0_mean'] = self.parameters['v0'].clone().detach()
            self.MCMC_toolbox['priors']['s_v0'] = 0.1
            # TODO? same on g?

        # TODO? why not passing the ready-to-use collection realizations that is initialized at beginning of fit algo and use it here instead?
        population_dictionary = self._create_dictionary_of_population_realizations()
        self.update_MCMC_toolbox(["all"], population_dictionary)

    def update_MCMC_toolbox(self, name_of_the_variables_that_have_been_changed, realizations):
        L = name_of_the_variables_that_have_been_changed
        values = {}
        if any(c in L for c in ('g', 'all')):
            values['g'] = realizations['g'].tensor_realizations
        if any(c in L for c in ('v0', 'v0_collinear', 'all')):
            values['v0'] = realizations['v0'].tensor_realizations
        if any(c in L for c in ('betas', 'all')) and self.source_dimension != 0:
            values['betas'] = realizations['betas'].tensor_realizations

        self.MCMC_toolbox['attributes'].update(name_of_the_variables_that_have_been_changed, values)

    def _center_xi_realizations(self, realizations):
        # This operation does not change the orthonormal basis
        # (since the resulting v0 is collinear to the previous one)
        # Nor all model computations (only v0 * exp(xi_i) matters),
        # it is only intended for model identifiability / `xi_i` regularization
        # <!> all operations are performed in "log" space (v0 is log'ed)
        mean_xi = torch.mean(realizations['xi'].tensor_realizations)
        realizations['xi'].tensor_realizations = realizations['xi'].tensor_realizations - mean_xi
        realizations['v0'].tensor_realizations = realizations['v0'].tensor_realizations + mean_xi

        self.update_MCMC_toolbox(['v0_collinear'], realizations)

        return realizations

    def compute_sufficient_statistics(self, data, realizations):

        # modify realizations in-place
        realizations = self._center_xi_realizations(realizations)

        # unlink all sufficient statistics from updates in realizations!
        realizations = realizations.clone_realizations()

        sufficient_statistics = {
            'g': realizations['g'].tensor_realizations,
            'v0': realizations['v0'].tensor_realizations,
            'tau': realizations['tau'].tensor_realizations,
            'tau_sqrd': torch.pow(realizations['tau'].tensor_realizations, 2),
            'xi': realizations['xi'].tensor_realizations,
            'xi_sqrd': torch.pow(realizations['xi'].tensor_realizations, 2)
        }
        if self.source_dimension != 0:
            sufficient_statistics['betas'] = realizations['betas'].tensor_realizations

        individual_parameters = self.get_param_from_real(realizations)

        data_reconstruction = self.compute_individual_tensorized(data.timepoints,
                                                                 individual_parameters,
                                                                 attribute_type='MCMC')

        data_reconstruction *= data.mask.float()  # speed-up computations

        norm_1 = data.values * data_reconstruction
        norm_2 = data_reconstruction * data_reconstruction

        sufficient_statistics['obs_x_reconstruction'] = norm_1  # .sum(dim=2) # no sum on features...
        sufficient_statistics['reconstruction_x_reconstruction'] = norm_2  # .sum(dim=2) # no sum on features...

        if self.noise_model == 'bernoulli':
            sufficient_statistics['crossentropy'] = self.compute_individual_attachment_tensorized(data, individual_parameters,
                                                                                                  attribute_type='MCMC')

        return sufficient_statistics

    def update_model_parameters_burn_in(self, data, realizations):
        # During the burn-in phase, we only need to store the following parameters (cf. !66 and #60)
        # - noise_std
        # - *_mean/std for regularization of individual variables
        # - others population parameters for regularization of population variables
        # We don't need to update the model "attributes" (never used during burn-in!)

        # TODO: refactorize?

        # modify realizations in-place!
        realizations = self._center_xi_realizations(realizations)

        # unlink model parameters from updates in realizations!
        realizations = realizations.clone_realizations()

        # Memoryless part of the algorithm
        self.parameters['g'] = realizations['g'].tensor_realizations

        v0_emp = realizations['v0'].tensor_realizations
        if self.MCMC_toolbox['priors'].get('v0_mean', None) is not None:
            v0_mean = self.MCMC_toolbox['priors']['v0_mean']
            s_v0 = self.MCMC_toolbox['priors']['s_v0']
            sigma_v0 = self.MCMC_toolbox['priors']['v0_std']
            self.parameters['v0'] = (1 / (1 / (s_v0 ** 2) + 1 / (sigma_v0 ** 2))) * (
                        v0_emp / (sigma_v0 ** 2) + v0_mean / (s_v0 ** 2))
        else:
            # new default
            self.parameters['v0'] = v0_emp

        if self.source_dimension != 0:
            self.parameters['betas'] = realizations['betas'].tensor_realizations

        xi = realizations['xi'].tensor_realizations
        # self.parameters['xi_mean'] = torch.mean(xi)  # fixed = 0 by design
        self.parameters['xi_std'] = torch.std(xi)
        tau = realizations['tau'].tensor_realizations
        self.parameters['tau_mean'] = torch.mean(tau)
        self.parameters['tau_std'] = torch.std(tau)

        # by design: sources_mean = 0., sources_std = 1.

        param_ind = self.get_param_from_real(realizations)
        self.parameters['noise_std'] = NoiseModel.rmse_model(self, data, param_ind, attribute_type='MCMC')

        if self.noise_model == 'bernoulli':
            self.parameters['crossentropy'] = self.compute_individual_attachment_tensorized(data, param_ind,
                                                                                            attribute_type='MCMC').sum()

    def update_model_parameters_normal(self, data, suff_stats):
        # TODO? add a true, configurable, validation for all parameters? (e.g.: bounds on tau_var/std but also on tau_mean, ...)

        # Stochastic sufficient statistics used to update the parameters of the model

        # TODO with Raphael : check the SS, especially the issue with mean(xi) and v_k
        # TODO : 1. Learn the mean of xi and v_k
        # TODO : 2. Set the mean of xi to 0 and add it to the mean of V_k
        self.parameters['g'] = suff_stats['g']
        self.parameters['v0'] = suff_stats['v0']
        if self.source_dimension != 0:
            self.parameters['betas'] = suff_stats['betas']

        tau_mean = self.parameters['tau_mean']
        tau_var_updt = torch.mean(suff_stats['tau_sqrd']) - 2. * tau_mean * torch.mean(suff_stats['tau'])
        tau_var = tau_var_updt + tau_mean ** 2
        self.parameters['tau_std'] = self._compute_std_from_var(tau_var, varname='tau_std')
        self.parameters['tau_mean'] = torch.mean(suff_stats['tau'])

        xi_mean = self.parameters['xi_mean']
        xi_var_updt = torch.mean(suff_stats['xi_sqrd']) - 2. * xi_mean * torch.mean(suff_stats['xi'])
        xi_var = xi_var_updt + xi_mean ** 2
        self.parameters['xi_std'] = self._compute_std_from_var(xi_var, varname='xi_std')
        # self.parameters['xi_mean'] = torch.mean(suff_stats['xi'])  # fixed = 0 by design

        if 'scalar' in self.noise_model:
            # scalar noise (same for all features)
            S1 = data.L2_norm
            S2 = suff_stats['obs_x_reconstruction'].sum()
            S3 = suff_stats['reconstruction_x_reconstruction'].sum()

            noise_var = (S1 - 2. * S2 + S3) / data.n_observations
        else:
            # keep feature dependence on feature to update diagonal noise (1 free param per feature)
            S1 = data.L2_norm_per_ft
            S2 = suff_stats['obs_x_reconstruction'].sum(dim=(0, 1))
            S3 = suff_stats['reconstruction_x_reconstruction'].sum(dim=(0, 1))

            # tensor 1D, shape (dimension,)
            noise_var = (S1 - 2. * S2 + S3) / data.n_observations_per_ft.float()

        self.parameters['noise_std'] = self._compute_std_from_var(noise_var, varname='noise_std')

        if self.noise_model == 'bernoulli':
            self.parameters['crossentropy'] = suff_stats['crossentropy'].sum()

    ###################################
    ### Random Variable Information ###
    ###################################

    def random_variable_informations(self):

        # --- Population variables
        g_infos = {
            "name": "g",
            "shape": torch.Size([self.dimension]),
            "type": "population",
            "rv_type": "multigaussian"
        }

        v0_infos = {
            "name": "v0",
            "shape": torch.Size([self.dimension]),
            "type": "population",
            "rv_type": "multigaussian"
        }

        betas_infos = {
            "name": "betas",
            "shape": torch.Size([self.dimension - 1, self.source_dimension]),
            "type": "population",
            "rv_type": "multigaussian",
            "scale": .5  # cf. GibbsSampler
        }

        # --- Individual variables
        tau_infos = {
            "name": "tau",
            "shape": torch.Size([1]),
            "type": "individual",
            "rv_type": "gaussian"
        }

        xi_infos = {
            "name": "xi",
            "shape": torch.Size([1]),
            "type": "individual",
            "rv_type": "gaussian"
        }

        sources_infos = {
            "name": "sources",
            "shape": torch.Size([self.source_dimension]),
            "type": "individual",
            "rv_type": "gaussian"
        }

        variables_infos = {
            "g": g_infos,
            "v0": v0_infos,
            "tau": tau_infos,
            "xi": xi_infos,
        }

        if self.source_dimension != 0:
            variables_infos['sources'] = sources_infos
            variables_infos['betas'] = betas_infos

        return variables_infos

# document some methods (we cannot decorate them at method creation since they are not yet decorated from `doc_with_super`)
doc_with_(MultivariateModel.compute_individual_tensorized_linear,
          MultivariateModel.compute_individual_tensorized,
          mapping={'the model': 'the model (linear)'})
doc_with_(MultivariateModel.compute_individual_tensorized_logistic,
          MultivariateModel.compute_individual_tensorized,
          mapping={'the model': 'the model (logistic)'})
#doc_with_(MultivariateModel.compute_individual_tensorized_mixed,
#          MultivariateModel.compute_individual_tensorized,
#          mapping={'the model': 'the model (mixed logistic-linear)'})

doc_with_(MultivariateModel.compute_jacobian_tensorized_linear,
          MultivariateModel.compute_jacobian_tensorized,
          mapping={'the model': 'the model (linear)'})
doc_with_(MultivariateModel.compute_jacobian_tensorized_logistic,
          MultivariateModel.compute_jacobian_tensorized,
          mapping={'the model': 'the model (logistic)'})
#doc_with_(MultivariateModel.compute_jacobian_tensorized_mixed,
#          MultivariateModel.compute_jacobian_tensorized,
#          mapping={'the model': 'the model (mixed logistic-linear)'})

#doc_with_(MultivariateModel.compute_individual_ages_from_biomarker_values_tensorized_linear,
#          MultivariateModel.compute_individual_ages_from_biomarker_values_tensorized,
#          mapping={'the model': 'the model (linear)'})
doc_with_(MultivariateModel.compute_individual_ages_from_biomarker_values_tensorized_logistic,
          MultivariateModel.compute_individual_ages_from_biomarker_values_tensorized,
          mapping={'the model': 'the model (logistic)'})
#doc_with_(MultivariateModel.compute_individual_ages_from_biomarker_values_tensorized_mixed,
#          MultivariateModel.compute_individual_ages_from_biomarker_values_tensorized,
#          mapping={'the model': 'the model (mixed logistic-linear)'})



import os
import ast
from setuptools import setup

def find_version(*py_file_with_version_paths):
    with open(os.path.join(*py_file_with_version_paths), 'r') as f:
        for line in f:
            if line.startswith('__version__'):
                return ast.parse(line).body[0].value.s # string
    raise RuntimeError("Unable to find version string.")

def readme():
    with open('README.md', 'r') as f:
        return f.read()

version = find_version("leaspy", "__init__.py")

with open("requirements.txt", 'r') as f:
    requirements = f.read().splitlines()

with open("requirements_dev.txt", 'r') as f:
    dev_requirements = f.read().splitlines()

with open("docs/requirements.txt", 'r') as f:
    docs_requirements = f.read().splitlines()

EXTRAS_REQUIRE = {
    'dev': dev_requirements,
    'docs': docs_requirements
}

setup(name="leaspy",
      version=version,

      description='Leaspy is a software package for the statistical analysis of longitudinal data.',
      long_description=readme(),
      long_description_content_type='text/markdown',
      license='BSD-3-Clause',

      url='https://gitlab.com/icm-institute/aramislab/leaspy',
      project_urls={
          'Bug Reports': 'https://gitlab.com/icm-institute/aramislab/leaspy/issues',
          'Source': 'https://gitlab.com/icm-institute/aramislab/leaspy',
          'Documentation': 'https://leaspy.readthedocs.io',
      },

      author='Igor Koval, Raphael Couronne, Etienne Maheux, Arnaud Valladier, Benoit Martin, Pierre-Emmanuel Poulet, Samuel Gruffaz, Cecile Di Folco, Juliette Ortholand, Mkrtich Vatinyan, Benoit Sauty De Chalon, Nemo Fournier, Quentin Madura, Stanley Durrleman',  # TODO
      author_email='igor.koval@inria.fr',
      maintainer='Igor Koval',
      maintainer_email='igor.koval@inria.fr',

      python_requires='>=3.7, <3.10',

      keywords='leaspy longitudinal mixed-model',

      packages=['leaspy',

                'leaspy.algo',
                'leaspy.algo.fit',
                'leaspy.algo.personalize',
                'leaspy.algo.simulate',
                'leaspy.algo.others',
                'leaspy.algo.utils',
                'leaspy.algo.utils.samplers',

                'leaspy.datasets',

                'leaspy.io',
                'leaspy.io.data',
                'leaspy.io.settings',
                'leaspy.io.realizations',
                'leaspy.io.outputs',
                'leaspy.io.logs',
                'leaspy.io.logs.visualization',

                'leaspy.models',
                'leaspy.models.utils',
                'leaspy.models.utils.attributes',
                'leaspy.models.utils.initialization',

                'leaspy.utils',
                ],

      install_requires=requirements,
      include_package_data=True,
      data_files=[('requirements', ['requirements.txt', 'requirements_dev.txt', 'docs/requirements.txt'])],

      # tests_require=["unittest"],
      test_suite='tests',

      classifiers=[
          # https://pypi.org/classifiers/
          "Development Status :: 5 - Production/Stable",
          "Intended Audience :: Science/Research",
          "Intended Audience :: Developers",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "License :: OSI Approved :: BSD License",
          "Operating System :: OS Independent",
      ],
      extras_require=EXTRAS_REQUIRE
    )
